# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Server-side pack repository related request implmentations."""

from .repository import SmartServerRepositoryRequest
from .request import SuccessfulSmartServerResponse


class SmartServerPackRepositoryAutopack(SmartServerRepositoryRequest):

    def do_repository_request(self, repository):
        pack_collection = getattr(repository, '_pack_collection', None)
        if pack_collection is None:
            # This is a not a pack repo, so asking for an autopack is just a
            # no-op.
            return SuccessfulSmartServerResponse((b'ok',))
        with repository.lock_write():
            repository._pack_collection.autopack()
        return SuccessfulSmartServerResponse((b'ok',))



# Copyright (C) 2007, 2008, 2009, 2011, 2012 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from breezy import branch as _mod_branch
from breezy import (controldir, errors, reconfigure, repository, tests,
                    workingtree)
from breezy.bzr import branch as _mod_bzrbranch
from breezy.bzr import vf_repository


class TestReconfigure(tests.TestCaseWithTransport):

    def test_tree_to_branch(self):
        tree = self.make_branch_and_tree('tree')
        reconfiguration = reconfigure.Reconfigure.to_branch(tree.controldir)
        reconfiguration.apply()
        self.assertRaises(errors.NoWorkingTree, workingtree.WorkingTree.open,
                          'tree')

    def test_modified_tree_to_branch(self):
        tree = self.make_branch_and_tree('tree')
        self.build_tree(['tree/file'])
        tree.add('file')
        reconfiguration = reconfigure.Reconfigure.to_branch(tree.controldir)
        self.assertRaises(errors.UncommittedChanges, reconfiguration.apply)
        reconfiguration.apply(force=True)
        self.assertRaises(errors.NoWorkingTree, workingtree.WorkingTree.open,
                          'tree')

    def test_tree_with_pending_merge_to_branch(self):
        tree = self.make_branch_and_tree('tree')
        tree.commit('unchanged')
        other_tree = tree.controldir.sprout('other').open_workingtree()
        other_tree.commit('mergeable commit')
        tree.merge_from_branch(other_tree.branch)
        reconfiguration = reconfigure.Reconfigure.to_branch(tree.controldir)
        self.assertRaises(errors.UncommittedChanges, reconfiguration.apply)
        reconfiguration.apply(force=True)
        self.assertRaises(errors.NoWorkingTree, workingtree.WorkingTree.open,
                          'tree')

    def test_branch_to_branch(self):
        branch = self.make_branch('branch')
        self.assertRaises(reconfigure.AlreadyBranch,
                          reconfigure.Reconfigure.to_branch, branch.controldir)

    def test_repo_to_branch(self):
        repo = self.make_repository('repo')
        reconfiguration = reconfigure.Reconfigure.to_branch(repo.controldir)
        reconfiguration.apply()

    def test_checkout_to_branch(self):
        branch = self.make_branch('branch')
        checkout = branch.create_checkout('checkout')
        reconfiguration = reconfigure.Reconfigure.to_branch(
            checkout.controldir)
        reconfiguration.apply()
        reconfigured = controldir.ControlDir.open('checkout').open_branch()
        self.assertIs(None, reconfigured.get_bound_location())

    def prepare_lightweight_checkout_to_branch(self):
        branch = self.make_branch('branch')
        checkout = branch.create_checkout('checkout', lightweight=True)
        checkout.commit('first commit', rev_id=b'rev1')
        reconfiguration = reconfigure.Reconfigure.to_branch(
            checkout.controldir)
        return reconfiguration, checkout

    def test_lightweight_checkout_to_branch(self):
        reconfiguration, checkout = \
            self.prepare_lightweight_checkout_to_branch()
        reconfiguration.apply()
        checkout_branch = checkout.controldir.open_branch()
        self.assertEqual(checkout_branch.controldir.root_transport.base,
                         checkout.controldir.root_transport.base)
        self.assertEqual(b'rev1', checkout_branch.last_revision())
        repo = checkout.controldir.open_repository()
        repo.get_revision(b'rev1')

    def test_lightweight_checkout_to_branch_tags(self):
        reconfiguration, checkout = \
            self.prepare_lightweight_checkout_to_branch()
        checkout.branch.tags.set_tag('foo', b'bar')
        reconfiguration.apply()
        checkout_branch = checkout.controldir.open_branch()
        self.assertEqual(b'bar', checkout_branch.tags.lookup_tag('foo'))

    def prepare_lightweight_checkout_to_checkout(self):
        branch = self.make_branch('branch')
        checkout = branch.create_checkout('checkout', lightweight=True)
        reconfiguration = reconfigure.Reconfigure.to_checkout(
            checkout.controldir)
        return reconfiguration, checkout

    def test_lightweight_checkout_to_checkout(self):
        reconfiguration, checkout = \
            self.prepare_lightweight_checkout_to_checkout()
        reconfiguration.apply()
        checkout_branch = checkout.controldir.open_branch()
        self.assertIsNot(checkout_branch.get_bound_location(), None)

    def test_lightweight_checkout_to_checkout_tags(self):
        reconfiguration, checkout = \
            self.prepare_lightweight_checkout_to_checkout()
        checkout.branch.tags.set_tag('foo', b'bar')
        reconfiguration.apply()
        checkout_branch = checkout.controldir.open_branch()
        self.assertEqual(b'bar', checkout_branch.tags.lookup_tag('foo'))

    def test_lightweight_conversion_uses_shared_repo(self):
        parent = self.make_branch('parent')
        shared_repo = self.make_repository('repo', shared=True)
        checkout = parent.create_checkout('repo/checkout', lightweight=True)
        reconfigure.Reconfigure.to_tree(checkout.controldir).apply()
        checkout_repo = checkout.controldir.open_branch().repository
        self.assertEqual(shared_repo.controldir.root_transport.base,
                         checkout_repo.controldir.root_transport.base)

    def test_branch_to_tree(self):
        branch = self.make_branch('branch')
        reconfiguration = reconfigure.Reconfigure.to_tree(branch.controldir)
        reconfiguration.apply()
        branch.controldir.open_workingtree()

    def test_tree_to_tree(self):
        tree = self.make_branch_and_tree('tree')
        self.assertRaises(reconfigure.AlreadyTree,
                          reconfigure.Reconfigure.to_tree, tree.controldir)

    def test_select_bind_location(self):
        branch = self.make_branch('branch')
        reconfiguration = reconfigure.Reconfigure(branch.controldir)
        self.assertRaises(reconfigure.NoBindLocation,
                          reconfiguration._select_bind_location)
        branch.set_parent('http://parent')
        reconfiguration = reconfigure.Reconfigure(branch.controldir)
        self.assertEqual('http://parent',
                         reconfiguration._select_bind_location())
        branch.set_push_location('sftp://push')
        reconfiguration = reconfigure.Reconfigure(branch.controldir)
        self.assertEqual('sftp://push',
                         reconfiguration._select_bind_location())
        branch.lock_write()
        try:
            branch.set_bound_location('bzr://foo/old-bound')
            branch.set_bound_location(None)
        finally:
            branch.unlock()
        reconfiguration = reconfigure.Reconfigure(branch.controldir)
        self.assertEqual('bzr://foo/old-bound',
                         reconfiguration._select_bind_location())
        branch.set_bound_location('bzr://foo/cur-bound')
        reconfiguration = reconfigure.Reconfigure(branch.controldir)
        self.assertEqual('bzr://foo/cur-bound',
                         reconfiguration._select_bind_location())
        reconfiguration.new_bound_location = 'ftp://user-specified'
        self.assertEqual('ftp://user-specified',
                         reconfiguration._select_bind_location())

    def test_select_reference_bind_location(self):
        branch = self.make_branch('branch')
        checkout = branch.create_checkout('checkout', lightweight=True)
        reconfiguration = reconfigure.Reconfigure(checkout.controldir)
        self.assertEqual(branch.base,
                         reconfiguration._select_bind_location())

    def test_tree_to_checkout(self):
        # A tree with no related branches and no supplied bind location cannot
        # become a checkout
        parent = self.make_branch('parent')

        tree = self.make_branch_and_tree('tree')
        reconfiguration = reconfigure.Reconfigure.to_checkout(tree.controldir)
        self.assertRaises(reconfigure.NoBindLocation, reconfiguration.apply)
        # setting a parent allows it to become a checkout
        tree.branch.set_parent(parent.base)
        reconfiguration = reconfigure.Reconfigure.to_checkout(tree.controldir)
        reconfiguration.apply()
        # supplying a location allows it to become a checkout
        tree2 = self.make_branch_and_tree('tree2')
        reconfiguration = reconfigure.Reconfigure.to_checkout(tree2.controldir,
                                                              parent.base)
        reconfiguration.apply()

    def test_tree_to_lightweight_checkout(self):
        # A tree with no related branches and no supplied bind location cannot
        # become a checkout
        parent = self.make_branch('parent')

        tree = self.make_branch_and_tree('tree')
        reconfiguration = reconfigure.Reconfigure.to_lightweight_checkout(
            tree.controldir)
        self.assertRaises(reconfigure.NoBindLocation, reconfiguration.apply)
        # setting a parent allows it to become a checkout
        tree.branch.set_parent(parent.base)
        reconfiguration = reconfigure.Reconfigure.to_lightweight_checkout(
            tree.controldir)
        reconfiguration.apply()
        # supplying a location allows it to become a checkout
        tree2 = self.make_branch_and_tree('tree2')
        reconfiguration = reconfigure.Reconfigure.to_lightweight_checkout(
            tree2.controldir, parent.base)
        reconfiguration.apply()

    def test_checkout_to_checkout(self):
        parent = self.make_branch('parent')
        checkout = parent.create_checkout('checkout')
        self.assertRaises(reconfigure.AlreadyCheckout,
                          reconfigure.Reconfigure.to_checkout, checkout.controldir)

    def make_unsynced_checkout(self):
        parent = self.make_branch('parent')
        checkout = parent.create_checkout('checkout')
        checkout.commit('test', rev_id=b'new-commit', local=True)
        reconfiguration = reconfigure.Reconfigure.to_lightweight_checkout(
            checkout.controldir)
        return checkout, parent, reconfiguration

    def test_unsynced_checkout_to_lightweight(self):
        checkout, parent, reconfiguration = self.make_unsynced_checkout()
        self.assertRaises(reconfigure.UnsyncedBranches, reconfiguration.apply)

    def test_synced_checkout_to_lightweight(self):
        checkout, parent, reconfiguration = self.make_unsynced_checkout()
        parent.pull(checkout.branch)
        reconfiguration.apply()
        wt = checkout.controldir.open_workingtree()
        self.assertTrue(parent.repository.has_same_location(
            wt.branch.repository))
        parent.repository.get_revision(b'new-commit')
        self.assertRaises(errors.NoRepositoryPresent,
                          checkout.controldir.open_repository)

    def prepare_branch_to_lightweight_checkout(self):
        parent = self.make_branch('parent')
        child = parent.controldir.sprout('child').open_workingtree()
        child.commit('test', rev_id=b'new-commit')
        parent.pull(child.branch)
        child.controldir.destroy_workingtree()
        reconfiguration = reconfigure.Reconfigure.to_lightweight_checkout(
            child.controldir)
        return parent, child, reconfiguration

    def test_branch_to_lightweight_checkout(self):
        parent, child, reconfiguration = \
            self.prepare_branch_to_lightweight_checkout()
        reconfiguration.apply()
        self.assertTrue(reconfiguration._destroy_branch)
        wt = child.controldir.open_workingtree()
        self.assertTrue(parent.repository.has_same_location(
            wt.branch.repository))
        parent.repository.get_revision(b'new-commit')
        self.assertRaises(errors.NoRepositoryPresent,
                          child.controldir.open_repository)

    def test_branch_to_lightweight_checkout_failure(self):
        parent, child, reconfiguration = \
            self.prepare_branch_to_lightweight_checkout()
        old_Repository_fetch = vf_repository.VersionedFileRepository.fetch
        vf_repository.VersionedFileRepository.fetch = None
        try:
            self.assertRaises(TypeError, reconfiguration.apply)
        finally:
            vf_repository.VersionedFileRepository.fetch = old_Repository_fetch
        child = _mod_branch.Branch.open('child')
        self.assertContainsRe(child.base, 'child/$')

    def test_branch_to_lightweight_checkout_fetch_tags(self):
        parent, child, reconfiguration = \
            self.prepare_branch_to_lightweight_checkout()
        child.branch.tags.set_tag('foo', b'bar')
        reconfiguration.apply()
        child = _mod_branch.Branch.open('child')
        self.assertEqual(b'bar', parent.tags.lookup_tag('foo'))

    def test_lightweight_checkout_to_lightweight_checkout(self):
        parent = self.make_branch('parent')
        checkout = parent.create_checkout('checkout', lightweight=True)
        self.assertRaises(reconfigure.AlreadyLightweightCheckout,
                          reconfigure.Reconfigure.to_lightweight_checkout,
                          checkout.controldir)

    def test_repo_to_tree(self):
        repo = self.make_repository('repo')
        reconfiguration = reconfigure.Reconfigure.to_tree(repo.controldir)
        reconfiguration.apply()
        workingtree.WorkingTree.open('repo')

    def test_shared_repo_to_lightweight_checkout(self):
        repo = self.make_repository('repo', shared=True)
        reconfiguration = reconfigure.Reconfigure.to_lightweight_checkout(
            repo.controldir)
        self.assertRaises(reconfigure.NoBindLocation, reconfiguration.apply)
        self.make_branch('branch')
        reconfiguration = reconfigure.Reconfigure.to_lightweight_checkout(
            repo.controldir, 'branch')
        reconfiguration.apply()
        workingtree.WorkingTree.open('repo')
        repository.Repository.open('repo')

    def test_unshared_repo_to_lightweight_checkout(self):
        repo = self.make_repository('repo', shared=False)
        self.make_branch('branch')
        reconfiguration = reconfigure.Reconfigure.to_lightweight_checkout(
            repo.controldir, 'branch')
        reconfiguration.apply()
        workingtree.WorkingTree.open('repo')
        self.assertRaises(errors.NoRepositoryPresent,
                          repository.Repository.open, 'repo')

    def test_standalone_to_use_shared(self):
        self.build_tree(['root/'])
        tree = self.make_branch_and_tree('root/tree')
        tree.commit('Hello', rev_id=b'hello-id')
        repo = self.make_repository('root', shared=True)
        reconfiguration = reconfigure.Reconfigure.to_use_shared(
            tree.controldir)
        reconfiguration.apply()
        tree = workingtree.WorkingTree.open('root/tree')
        self.assertTrue(repo.has_same_location(tree.branch.repository))
        self.assertEqual('Hello', repo.get_revision(b'hello-id').message)

    def add_dead_head(self, tree):
        revno, revision_id = tree.branch.last_revision_info()
        tree.commit('Dead head', rev_id=b'dead-head-id')
        tree.branch.set_last_revision_info(revno, revision_id)
        tree.set_last_revision(revision_id)

    def test_standalone_to_use_shared_preserves_dead_heads(self):
        self.build_tree(['root/'])
        tree = self.make_branch_and_tree('root/tree')
        self.add_dead_head(tree)
        tree.commit('Hello', rev_id=b'hello-id')
        repo = self.make_repository('root', shared=True)
        reconfiguration = reconfigure.Reconfigure.to_use_shared(
            tree.controldir)
        reconfiguration.apply()
        tree = workingtree.WorkingTree.open('root/tree')
        message = repo.get_revision(b'dead-head-id').message
        self.assertEqual('Dead head', message)

    def make_repository_tree(self):
        self.build_tree(['root/'])
        self.make_repository('root', shared=True)
        tree = self.make_branch_and_tree('root/tree')
        reconfigure.Reconfigure.to_use_shared(tree.controldir).apply()
        return workingtree.WorkingTree.open('root/tree')

    def test_use_shared_to_use_shared(self):
        tree = self.make_repository_tree()
        self.assertRaises(reconfigure.AlreadyUsingShared,
                          reconfigure.Reconfigure.to_use_shared, tree.controldir)

    def test_use_shared_to_standalone(self):
        tree = self.make_repository_tree()
        tree.commit('Hello', rev_id=b'hello-id')
        reconfigure.Reconfigure.to_standalone(tree.controldir).apply()
        tree = workingtree.WorkingTree.open('root/tree')
        repo = tree.branch.repository
        self.assertEqual(repo.controldir.root_transport.base,
                         tree.controldir.root_transport.base)
        self.assertEqual('Hello', repo.get_revision(b'hello-id').message)

    def test_use_shared_to_standalone_preserves_dead_heads(self):
        tree = self.make_repository_tree()
        self.add_dead_head(tree)
        tree.commit('Hello', rev_id=b'hello-id')
        reconfigure.Reconfigure.to_standalone(tree.controldir).apply()
        tree = workingtree.WorkingTree.open('root/tree')
        repo = tree.branch.repository
        self.assertRaises(errors.NoSuchRevision, repo.get_revision,
                          b'dead-head-id')

    def test_standalone_to_standalone(self):
        tree = self.make_branch_and_tree('tree')
        self.assertRaises(reconfigure.AlreadyStandalone,
                          reconfigure.Reconfigure.to_standalone, tree.controldir)

    def make_unsynced_branch_reconfiguration(self):
        parent = self.make_branch_and_tree('parent')
        parent.commit('commit 1')
        child = parent.controldir.sprout('child').open_workingtree()
        child.commit('commit 2')
        return reconfigure.Reconfigure.to_lightweight_checkout(child.controldir)

    def test_unsynced_branch_to_lightweight_checkout_unforced(self):
        reconfiguration = self.make_unsynced_branch_reconfiguration()
        self.assertRaises(reconfigure.UnsyncedBranches, reconfiguration.apply)

    def test_unsynced_branch_to_lightweight_checkout_forced(self):
        reconfiguration = self.make_unsynced_branch_reconfiguration()
        reconfiguration.apply(force=True)

    def make_repository_with_without_trees(self, with_trees):
        repo = self.make_repository('repo', shared=True)
        repo.set_make_working_trees(with_trees)
        return repo

    def test_make_with_trees(self):
        repo = self.make_repository_with_without_trees(False)
        reconfiguration = reconfigure.Reconfigure.set_repository_trees(
            repo.controldir, True)
        reconfiguration.apply()
        self.assertIs(True, repo.make_working_trees())

    def test_make_without_trees(self):
        repo = self.make_repository_with_without_trees(True)
        reconfiguration = reconfigure.Reconfigure.set_repository_trees(
            repo.controldir, False)
        reconfiguration.apply()
        self.assertIs(False, repo.make_working_trees())

    def test_make_with_trees_already_with_trees(self):
        repo = self.make_repository_with_without_trees(True)
        e = self.assertRaises(reconfigure.AlreadyWithTrees,
                              reconfigure.Reconfigure.set_repository_trees, repo.controldir, True)
        self.assertContainsRe(str(e),
                              r"Shared repository '.*' already creates working trees.")

    def test_make_without_trees_already_no_trees(self):
        repo = self.make_repository_with_without_trees(False)
        e = self.assertRaises(
            reconfigure.AlreadyWithNoTrees,
            reconfigure.Reconfigure.set_repository_trees, repo.controldir,
            False)
        self.assertContainsRe(
            str(e),
            r"Shared repository '.*' already doesn't create working trees.")

    def test_repository_tree_reconfiguration_not_supported(self):
        tree = self.make_branch_and_tree('tree')
        e = self.assertRaises(
            reconfigure.ReconfigurationNotSupported,
            reconfigure.Reconfigure.set_repository_trees, tree.controldir,
            None)
        self.assertContainsRe(
            str(e), r"Requested reconfiguration of '.*' is not supported.")

    def test_lightweight_checkout_to_tree_preserves_reference_locations(self):
        format = controldir.format_registry.make_controldir('1.9')
        format.set_branch_format(_mod_bzrbranch.BzrBranchFormat8())
        tree = self.make_branch_and_tree('tree', format=format)
        tree.branch.set_reference_info(b'file_id', '../location', 'path')
        checkout = tree.branch.create_checkout('checkout', lightweight=True)
        reconfiguration = reconfigure.Reconfigure.to_tree(checkout.controldir)
        reconfiguration.apply()
        checkout_branch = checkout.controldir.open_branch()
        self.assertEqual(('../location', 'path'),
                         checkout_branch.get_reference_info(b'file_id'))



# Copyright (C) 2005-2012, 2016 Canonical Ltd
# Copyright (C) 2017-2018 Breezy developers
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Blackbox tests for plugins behaviour and commands."""

import os

from ... import commands
from ..test_plugins import BaseTestPlugins


class TestPluginHelp(BaseTestPlugins):

    def run_bzr_utf8_out(self, *args, **kwargs):
        out, _ = self.run_bzr(*args, **kwargs)
        return out

    def split_help_commands(self):
        help = {}
        current = None
        out = self.run_bzr_utf8_out('--no-plugins help commands')
        for line in out.splitlines():
            if not line.startswith(' '):
                current = line.split()[0]
            help[current] = help.get(current, '') + line

        return help

    def test_plugin_help_builtins_unaffected(self):
        # Check we don't get false positives
        help_commands = self.split_help_commands()
        for cmd_name in commands.builtin_command_names():
            if cmd_name in commands.plugin_command_names():
                continue
            try:
                help = commands.get_cmd_object(cmd_name).get_help_text()
            except NotImplementedError:
                # some commands have no help
                pass
            else:
                self.assertNotContainsRe(help, 'plugin "[^"]*"')

            if cmd_name in help_commands:
                # some commands are hidden
                help = help_commands[cmd_name]
                self.assertNotContainsRe(help, 'plugin "[^"]*"')

    def test_plugin_help_shows_plugin(self):
        # Create a test plugin
        os.mkdir('plugin_test')
        source = (
            "from breezy import commands\n"
            "class cmd_myplug(commands.Command):\n"
            "    __doc__ = '''Just a simple test plugin.'''\n"
            "    aliases = ['mplg']\n"
            "    def run(self):\n"
            "        print ('Hello from my plugin')\n"
        )
        self.create_plugin('myplug', source, 'plugin_test')

        # Check its help
        self.load_with_paths(['plugin_test'])
        myplug = self.plugins['myplug'].module
        commands.register_command(myplug.cmd_myplug)
        self.addCleanup(commands.plugin_cmds.remove, 'myplug')
        help = self.run_bzr_utf8_out('help myplug')
        self.assertContainsRe(help, 'plugin "myplug"')
        help = self.split_help_commands()['myplug']
        self.assertContainsRe(help, '\\[myplug\\]')



# Copyright (C) 2006-2010 Canonical Ltd
# Authors: Robert Collins <robert.collins@canonical.com>
# -*- coding: utf-8 -*-
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""ControlDir implementation tests for bzr.

These test the conformance of all the controldir variations to the expected API.
Specific tests for individual formats are in the tests/test_bzrdir.py file
rather than in tests/per_branch/*.py.
"""

from breezy.controldir import ControlDirFormat
from breezy.tests import (TestCaseWithTransport, default_transport,
                          multiply_tests, test_server)
from breezy.transport import memory


def make_scenarios(vfs_factory, transport_server, transport_readonly_server,
                   formats, name_suffix=''):
    """Transform the input to a list of scenarios.

    :param formats: A list of bzrdir_format objects.
    :param vfs_server: A factory to create a Transport Server which has
        all the VFS methods working, and is writable.
    """
    result = []
    for format in formats:
        scenario_name = format.__class__.__name__
        scenario_name += name_suffix
        scenario = (scenario_name, {
            "vfs_transport_factory": vfs_factory,
            "transport_server": transport_server,
            "transport_readonly_server": transport_readonly_server,
            "bzrdir_format": format,
            })
        result.append(scenario)
    return result


class TestCaseWithControlDir(TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        self.controldir = None

    def get_bzrdir(self):
        if self.controldir is None:
            self.controldir = self.make_controldir(None)
        return self.controldir

    def get_default_format(self):
        return self.bzrdir_format


def load_tests(loader, standard_tests, pattern):
    test_per_controldir = [
        'breezy.tests.per_controldir.test_controldir',
        'breezy.tests.per_controldir.test_format',
        'breezy.tests.per_controldir.test_push',
        ]
    submod_tests = loader.loadTestsFromModuleNames(test_per_controldir)
    formats = ControlDirFormat.known_formats()
    scenarios = make_scenarios(
        default_transport,
        None,
        # None here will cause a readonly decorator to be created
        # by the TestCaseWithTransport.get_readonly_transport method.
        None,
        formats)
    # This will always add scenarios using the smart server.
    from breezy.bzr.remote import RemoteBzrDirFormat

    # test the remote server behaviour when backed with a MemoryTransport
    # Once for the current version
    scenarios.extend(make_scenarios(
        memory.MemoryServer,
        test_server.SmartTCPServer_for_testing,
        test_server.ReadonlySmartTCPServer_for_testing,
        [(RemoteBzrDirFormat())],
        name_suffix='-default'))
    # And once with < 1.6 - the 'v2' protocol.
    scenarios.extend(make_scenarios(
        memory.MemoryServer,
        test_server.SmartTCPServer_for_testing_v2_only,
        test_server.ReadonlySmartTCPServer_for_testing_v2_only,
        [(RemoteBzrDirFormat())],
        name_suffix='-v2'))
    # add the tests for the sub modules
    return multiply_tests(submod_tests, scenarios, standard_tests)



# Copyright (C) 2010, 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Test symlink support.
"""

import os

from breezy import osutils, tests, workingtree
from breezy.tests import features
from breezy.tests.per_workingtree import TestCaseWithWorkingTree


class TestSmartAddTree(TestCaseWithWorkingTree):

    # See eg <https://bugs.launchpad.net/bzr/+bug/192859>

    def setUp(self):
        super().setUp()
        self.requireFeature(features.SymlinkFeature(self.test_dir))

    def test_smart_add_symlink(self):
        tree = self.make_branch_and_tree('tree')
        self.build_tree_contents([
            ('tree/link@', b'target'),
            ])
        tree.smart_add(['tree/link'])
        self.assertTrue(tree.is_versioned('link'))
        self.assertFalse(tree.is_versioned('target'))
        self.assertEqual('symlink', tree.kind('link'))

    def test_smart_add_symlink_pointing_outside(self):
        tree = self.make_branch_and_tree('tree')
        self.build_tree_contents([
            ('tree/link@', '../../../../target'),
            ])
        tree.smart_add(['tree/link'])
        self.assertTrue(tree.is_versioned('link'))
        self.assertFalse(tree.is_versioned('target'))
        self.assertEqual('symlink', tree.kind('link'))

    def test_add_file_under_symlink(self):
        # similar to
        # https://bugs.launchpad.net/bzr/+bug/192859/comments/3
        tree = self.make_branch_and_tree('tree')
        self.build_tree_contents([
            ('tree/link@', 'dir'),
            ('tree/dir/',),
            ('tree/dir/file', b'content'),
            ])
        if tree.has_versioned_directories():
            self.assertEqual(
                tree.smart_add(['tree/link/file']),
                (['dir', 'dir/file'], {}))
        else:
            self.assertEqual(
                tree.smart_add(['tree/link/file']),
                (['dir/file'], {}))

        # should add the actual parent directory, not the apparent parent
        # (which is actually a symlink)
        self.assertTrue(tree.is_versioned('dir/file'))
        self.assertTrue(tree.is_versioned('dir'))
        self.assertFalse(tree.is_versioned('link'))
        self.assertFalse(tree.is_versioned('link/file'))


class TestKindChanges(TestCaseWithWorkingTree):

    def setUp(self):
        super().setUp()
        self.requireFeature(features.SymlinkFeature(self.test_dir))

    def test_symlink_changes_to_dir(self):
        # <https://bugs.launchpad.net/bzr/+bug/192859>:
        # we had some past problems with the workingtree remembering for too
        # long what kind of object was at a particular name; we really
        # shouldn't do that.  Operating on the dirstate through passing
        # inventory deltas rather than mutating the inventory largely avoids
        # that.
        tree = self.make_branch_and_tree('tree')
        self.build_tree_contents([
            ('tree/a@', 'target')])
        tree.smart_add(['tree/a'])
        tree.commit('add symlink')
        os.unlink('tree/a')
        self.build_tree_contents([
            ('tree/a/',),
            ('tree/a/f', b'content'),
            ])
        tree.smart_add(['tree/a/f'])
        tree.commit('change to dir')
        tree.lock_read()
        self.addCleanup(tree.unlock)
        self.assertEqual([], list(tree.iter_changes(tree.basis_tree())))
        self.assertEqual(
            ['a', 'a/f'], sorted(info[0] for info in tree.list_files()))

    def test_dir_changes_to_symlink(self):
        # <https://bugs.launchpad.net/bzr/+bug/192859>:
        # we had some past problems with the workingtree remembering for too
        # long what kind of object was at a particular name; we really
        # shouldn't do that.  Operating on the dirstate through passing
        # inventory deltas rather than mutating the inventory largely avoids
        # that.
        tree = self.make_branch_and_tree('tree')
        self.build_tree_contents([
            ('tree/a/',),
            ('tree/a/file', b'content'),
            ])
        tree.smart_add(['tree/a'])
        tree.commit('add dir')
        osutils.rmtree('tree/a')
        self.build_tree_contents([
            ('tree/a@', 'target'),
            ])
        tree.commit('change to symlink')


class TestOpenTree(TestCaseWithWorkingTree):

    def setUp(self):
        super().setUp()
        self.requireFeature(features.SymlinkFeature(self.test_dir))

    def test_open_containing_through_symlink(self):
        self.make_test_tree()
        self.check_open_containing('link/content', 'tree', 'content')
        self.check_open_containing('link/sublink', 'tree', 'sublink')
        # this next one is a bit debatable, but arguably it's better that
        # open_containing is only concerned with opening the tree
        # and then you can deal with symlinks along the way if you want
        self.check_open_containing('link/sublink/subcontent', 'tree',
                                   'sublink/subcontent')

    def check_open_containing(self, to_open, expected_tree_name,
                              expected_relpath):
        wt, relpath = workingtree.WorkingTree.open_containing(to_open)
        self.assertEqual(relpath, expected_relpath)
        self.assertEndsWith(wt.basedir, expected_tree_name)

    def test_tree_files(self):
        # not strictly a WorkingTree method, but it should be
        # probably the root cause for
        # <https://bugs.launchpad.net/bzr/+bug/128562>
        self.make_test_tree()
        self.check_tree_files(['tree/outerlink'],
                              'tree', ['outerlink'])
        self.check_tree_files(['link/outerlink'],
                              'tree', ['outerlink'])
        self.check_tree_files(['link/sublink/subcontent'],
                              'tree', ['subdir/subcontent'])

    def check_tree_files(self, to_open, expected_tree, expect_paths):
        tree, relpaths = workingtree.WorkingTree.open_containing_paths(to_open)
        self.assertEndsWith(tree.basedir, expected_tree)
        self.assertEqual(expect_paths, relpaths)

    def make_test_tree(self):
        tree = self.make_branch_and_tree('tree')
        self.build_tree_contents([
            ('link@', 'tree'),
            ('tree/outerlink@', '/not/there'),
            ('tree/content', b'hello'),
            ('tree/sublink@', 'subdir'),
            ('tree/subdir/',),
            ('tree/subdir/subcontent', b'subcontent stuff')
            ])



# Copyright (C) 2005 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

# \subsection{\emph{rio} - simple text metaformat}
#
# \emph{r} stands for `restricted', `reproducible', or `rfc822-like'.
#
# The stored data consists of a series of \emph{stanzas}, each of which contains
# \emph{fields} identified by an ascii name, with Unicode or string contents.
# The field tag is constrained to alphanumeric characters.
# There may be more than one field in a stanza with the same name.
#
# The format itself does not deal with character encoding issues, though
# the result will normally be written in Unicode.
#
# The format is intended to be simple enough that there is exactly one character
# stream representation of an object and vice versa, and that this relation
# will continue to hold for future versions of bzr.

import re

from .. import osutils
from ..iterablefile import IterableFile

# XXX: some redundancy is allowing to write stanzas in isolation as well as
# through a writer object.


class RioWriter:

    def __init__(self, to_file):
        self._soft_nl = False
        self._to_file = to_file

    def write_stanza(self, stanza):
        if self._soft_nl:
            self._to_file.write(b'\n')
        stanza.write(self._to_file)
        self._soft_nl = True


class RioReader:
    """Read stanzas from a file as a sequence

    to_file can be anything that can be enumerated as a sequence of
    lines (with newlines.)
    """

    def __init__(self, from_file):
        self._from_file = from_file

    def __iter__(self):
        while True:
            s = read_stanza(self._from_file)
            if s is None:
                break
            else:
                yield s


def rio_file(stanzas, header=None):
    """Produce a rio IterableFile from an iterable of stanzas"""
    def str_iter():
        if header is not None:
            yield header + b'\n'
        first_stanza = True
        for s in stanzas:
            if first_stanza is not True:
                yield b'\n'
            yield from s.to_lines()
            first_stanza = False
    return IterableFile(str_iter())


def read_stanzas(from_file):

    while True:
        s = read_stanza(from_file)
        if s is None:
            break
        yield s


class Stanza:
    """One stanza for rio.

    Each stanza contains a set of named fields.

    Names must be non-empty ascii alphanumeric plus _.  Names can be repeated
    within a stanza.  Names are case-sensitive.  The ordering of fields is
    preserved.

    Each field value must be either an int or a string.
    """

    __slots__ = ['items']

    def __init__(self, **kwargs):
        """Construct a new Stanza.

        The keyword arguments, if any, are added in sorted order to the stanza.
        """
        self.items = []
        if kwargs:
            for tag, value in sorted(kwargs.items()):
                self.add(tag, value)

    def add(self, tag, value):
        """Append a name and value to the stanza."""
        if not valid_tag(tag):
            raise ValueError("invalid tag {!r}".format(tag))
        if isinstance(value, bytes):
            pass
        elif isinstance(value, str):
            pass
        elif isinstance(value, Stanza):
            pass
        else:
            raise TypeError("invalid type for rio value: %r of type %s"
                            % (value, type(value)))
        self.items.append((tag, value))

    @classmethod
    def from_pairs(cls, pairs):
        ret = cls()
        ret.items = pairs
        return ret

    def __contains__(self, find_tag):
        """True if there is any field in this stanza with the given tag."""
        for tag, value in self.items:
            if tag == find_tag:
                return True
        return False

    def __len__(self):
        """Return number of pairs in the stanza."""
        return len(self.items)

    def __eq__(self, other):
        if not isinstance(other, Stanza):
            return False
        return self.items == other.items

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return "Stanza(%r)" % self.items

    def iter_pairs(self):
        """Return iterator of tag, value pairs."""
        return iter(self.items)

    def to_lines(self):
        """Generate sequence of lines for external version of this file.

        The lines are always utf-8 encoded strings.
        """
        if not self.items:
            # max() complains if sequence is empty
            return []
        result = []
        for text_tag, text_value in self.items:
            tag = text_tag.encode('ascii')
            if isinstance(text_value, str):
                value = text_value.encode('utf-8', 'surrogateescape')
            elif isinstance(text_value, Stanza):
                value = text_value.to_string()
            else:
                value = text_value
            if value == b'':
                result.append(tag + b': \n')
            elif b'\n' in value:
                # don't want splitlines behaviour on empty lines
                val_lines = value.split(b'\n')
                result.append(tag + b': ' + val_lines[0] + b'\n')
                for line in val_lines[1:]:
                    result.append(b'\t' + line + b'\n')
            else:
                result.append(tag + b': ' + value + b'\n')
        return result

    def to_string(self):
        """Return stanza as a single string"""
        return b''.join(self.to_lines())

    def write(self, to_file):
        """Write stanza to a file"""
        to_file.writelines(self.to_lines())

    def get(self, tag):
        """Return the value for a field wih given tag.

        If there is more than one value, only the first is returned.  If the
        tag is not present, KeyError is raised.
        """
        for t, v in self.items:
            if t == tag:
                return v
        else:
            raise KeyError(tag)

    __getitem__ = get

    def get_all(self, tag):
        r = []
        for t, v in self.items:
            if t == tag:
                r.append(v)
        return r

    def as_dict(self):
        """Return a dict containing the unique values of the stanza.
        """
        d = {}
        for tag, value in self.items:
            d[tag] = value
        return d


def valid_tag(tag):
    return _valid_tag(tag)


def read_stanza(line_iter):
    """Return new Stanza read from list of lines or a file

    Returns one Stanza that was read, or returns None at end of file.  If a
    blank line follows the stanza, it is consumed.  It's not an error for
    there to be no blank at end of file.  If there is a blank file at the
    start of the input this is really an empty stanza and that is returned.

    Only the stanza lines and the trailing blank (if any) are consumed
    from the line_iter.

    The raw lines must be in utf-8 encoding.
    """
    return _read_stanza_utf8(line_iter)


def to_patch_lines(stanza, max_width=72):
    """Convert a stanza into RIO-Patch format lines.

    RIO-Patch is a RIO variant designed to be e-mailed as part of a patch.
    It resists common forms of damage such as newline conversion or the removal
    of trailing whitespace, yet is also reasonably easy to read.

    :param max_width: The maximum number of characters per physical line.
    :return: a list of lines
    """
    if max_width <= 6:
        raise ValueError(max_width)
    max_rio_width = max_width - 4
    lines = []
    for pline in stanza.to_lines():
        for line in pline.split(b'\n')[:-1]:
            line = re.sub(b'\\\\', b'\\\\\\\\', line)
            while len(line) > 0:
                partline = line[:max_rio_width]
                line = line[max_rio_width:]
                if len(line) > 0 and line[:1] != [b' ']:
                    break_index = -1
                    break_index = partline.rfind(b' ', -20)
                    if break_index < 3:
                        break_index = partline.rfind(b'-', -20)
                        break_index += 1
                    if break_index < 3:
                        break_index = partline.rfind(b'/', -20)
                    if break_index >= 3:
                        line = partline[break_index:] + line
                        partline = partline[:break_index]
                if len(line) > 0:
                    line = b'  ' + line
                partline = re.sub(b'\r', b'\\\\r', partline)
                blank_line = False
                if len(line) > 0:
                    partline += b'\\'
                elif re.search(b' $', partline):
                    partline += b'\\'
                    blank_line = True
                lines.append(b'# ' + partline + b'\n')
                if blank_line:
                    lines.append(b'#   \n')
    return lines


def _patch_stanza_iter(line_iter):
    map = {b'\\\\': b'\\',
           b'\\r': b'\r',
           b'\\\n': b''}

    def mapget(match):
        return map[match.group(0)]

    last_line = None
    for line in line_iter:
        if line.startswith(b'# '):
            line = line[2:]
        elif line.startswith(b'#'):
            line = line[1:]
        else:
            raise ValueError("bad line {!r}".format(line))
        if last_line is not None and len(line) > 2:
            line = line[2:]
        line = re.sub(b'\r', b'', line)
        line = re.sub(b'\\\\(.|\n)', mapget, line)
        if last_line is None:
            last_line = line
        else:
            last_line += line
        if last_line[-1:] == b'\n':
            yield last_line
            last_line = None
    if last_line is not None:
        yield last_line


def read_patch_stanza(line_iter):
    """Convert an iterable of RIO-Patch format lines into a Stanza.

    RIO-Patch is a RIO variant designed to be e-mailed as part of a patch.
    It resists common forms of damage such as newline conversion or the removal
    of trailing whitespace, yet is also reasonably easy to read.

    :return: a Stanza
    """
    return read_stanza(_patch_stanza_iter(line_iter))


try:
    from ._rio_pyx import _read_stanza_utf8
    from ._rio_rs import _valid_tag
except ImportError as e:
    osutils.failed_to_load_extension(e)
    from ._rio_py import _read_stanza_utf8, _valid_tag



# Copyright (C) 2008 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Implementation of Transport that never has a smart medium.

This is mainly useful with HTTP transports, which sometimes have a smart medium
and sometimes don't.  By using this decorator, you can force those transports
to never have a smart medium.
"""

from .. import errors
from ..transport import decorator


class NoSmartTransportDecorator(decorator.TransportDecorator):
    """A decorator for transports that disables get_smart_medium."""

    @classmethod
    def _get_url_prefix(self):
        return 'nosmart+'

    def get_smart_medium(self):
        raise errors.NoSmartMedium(self)


def get_test_permutations():
    """Return the permutations to be used in testing."""
    from ..tests import test_server
    return [(NoSmartTransportDecorator, test_server.NoSmartTransportServer)]



# Copyright (C) 2005-2011, 2015, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for Transport implementations.

Transport implementations tested here are supplied by
TransportTestProviderAdapter.
"""

import os
import stat
import sys
from io import BytesIO

from .. import errors, osutils, pyutils, tests
from .. import transport as _mod_transport
from .. import urlutils
from ..errors import ConnectionError, PathError, TransportNotPossible
from ..osutils import getcwd
from ..transport import (ConnectedTransport, FileExists, NoSuchFile, Transport,
                         _get_transport_modules)
from ..transport.memory import MemoryTransport
from ..transport.remote import RemoteTransport
from . import TestNotApplicable, TestSkipped, multiply_tests, test_server
from .test_transport import TestTransportImplementation


def get_transport_test_permutations(module):
    """Get the permutations module wants to have tested."""
    if getattr(module, 'get_test_permutations', None) is None:
        raise AssertionError(
            "transport module %s doesn't provide get_test_permutations()"
            % module.__name__)
        return []
    return module.get_test_permutations()


def transport_test_permutations():
    """Return a list of the klass, server_factory pairs to test."""
    result = []
    for module in _get_transport_modules():
        try:
            permutations = get_transport_test_permutations(
                pyutils.get_named_object(module))
            for (klass, server_factory) in permutations:
                scenario = ('{},{}'.format(klass.__name__, server_factory.__name__),
                            {"transport_class": klass,
                             "transport_server": server_factory})
                result.append(scenario)
        except errors.DependencyNotPresent as e:
            # Continue even if a dependency prevents us
            # from adding this test
            pass
    return result


def load_tests(loader, standard_tests, pattern):
    """Multiply tests for tranport implementations."""
    result = loader.suiteClass()
    scenarios = transport_test_permutations()
    return multiply_tests(standard_tests, scenarios, result)


class TransportTests(TestTransportImplementation):

    def setUp(self):
        super().setUp()
        self.overrideEnv('BRZ_NO_SMART_VFS', None)

    def check_transport_contents(self, content, transport, relpath):
        """Check that transport.get_bytes(relpath) == content."""
        self.assertEqualDiff(content, transport.get_bytes(relpath))

    def test_ensure_base_missing(self):
        """.ensure_base() should create the directory if it doesn't exist"""
        t = self.get_transport()
        t_a = t.clone('a')
        self.assertFalse(t.ensure_base())
        if t_a.is_readonly():
            self.assertRaises(TransportNotPossible,
                              t_a.ensure_base)
            return
        self.assertTrue(t_a.ensure_base())
        self.assertTrue(t.has('a'))

    def test_ensure_base_exists(self):
        """.ensure_base() should just be happy if it already exists"""
        t = self.get_transport()
        if t.is_readonly():
            return

        t.mkdir('a')
        t_a = t.clone('a')
        # ensure_base returns False if it didn't create the base
        self.assertFalse(t_a.ensure_base())

    def test_ensure_base_missing_parent(self):
        """.ensure_base() will fail if the parent dir doesn't exist"""
        t = self.get_transport()
        if t.is_readonly():
            return

        t_a = t.clone('a')
        t_b = t_a.clone('b')
        self.assertRaises(NoSuchFile, t_b.ensure_base)

    def test_external_url(self):
        """.external_url either works or raises InProcessTransport."""
        t = self.get_transport()
        try:
            t.external_url()
        except errors.InProcessTransport:
            pass

    def test_has(self):
        t = self.get_transport()

        files = ['a', 'b', 'e', 'g', '%']
        self.build_tree(files, transport=t)
        self.assertEqual(True, t.has('a'))
        self.assertEqual(False, t.has('c'))
        self.assertEqual(True, t.has(urlutils.escape('%')))
        self.assertEqual(True, t.has_any(['a', 'b', 'c']))
        self.assertEqual(False, t.has_any(['c', 'd', 'f',
                                           urlutils.escape('%%')]))
        self.assertEqual(False, t.has_any(['c', 'c', 'c']))
        self.assertEqual(True, t.has_any(['b', 'b', 'b']))

    def test_has_root_works(self):
        if self.transport_server is test_server.SmartTCPServer_for_testing:
            raise TestNotApplicable(
                "SmartTCPServer_for_testing intentionally does not allow "
                "access to /.")
        current_transport = self.get_transport()
        self.assertTrue(current_transport.has('/'))
        root = current_transport.clone('/')
        self.assertTrue(root.has(''))

    def test_get(self):
        t = self.get_transport()

        files = ['a']
        content = b'contents of a\n'
        self.build_tree(['a'], transport=t, line_endings='binary')
        self.check_transport_contents(b'contents of a\n', t, 'a')
        f = t.get('a')
        self.assertEqual(content, f.read())

    def test_get_unknown_file(self):
        t = self.get_transport()
        files = ['a', 'b']
        contents = [b'contents of a\n',
                    b'contents of b\n',
                    ]
        self.build_tree(files, transport=t, line_endings='binary')
        self.assertRaises(NoSuchFile, t.get, 'c')

        def iterate_and_close(func, *args):
            for f in func(*args):
                # We call f.read() here because things like paramiko actually
                # spawn a thread to prefetch the content, which we want to
                # consume before we close the handle.
                content = f.read()
                f.close()

    def test_get_directory_read_gives_ReadError(self):
        """consistent errors for read() on a file returned by get()."""
        t = self.get_transport()
        if t.is_readonly():
            self.build_tree(['a directory/'])
        else:
            t.mkdir('a%20directory')
        # getting the file must either work or fail with a PathError
        try:
            a_file = t.get('a%20directory')
        except (errors.PathError, errors.RedirectRequested):
            # early failure return immediately.
            return
        # having got a file, read() must either work (i.e. http reading a dir
        # listing) or fail with ReadError
        try:
            a_file.read()
        except errors.ReadError:
            pass

    def test_get_bytes(self):
        t = self.get_transport()

        files = ['a', 'b', 'e', 'g']
        contents = [b'contents of a\n',
                    b'contents of b\n',
                    b'contents of e\n',
                    b'contents of g\n',
                    ]
        self.build_tree(files, transport=t, line_endings='binary')
        self.check_transport_contents(b'contents of a\n', t, 'a')

        for content, fname in zip(contents, files):
            self.assertEqual(content, t.get_bytes(fname))

    def test_get_bytes_unknown_file(self):
        t = self.get_transport()
        self.assertRaises(NoSuchFile, t.get_bytes, 'c')

    def test_get_bytes_with_open_write_stream_sees_all_content(self):
        t = self.get_transport()
        if t.is_readonly():
            return
        with t.open_write_stream('foo') as handle:
            handle.write(b'b')
            self.assertEqual(b'b', t.get_bytes('foo'))
            with t.get('foo') as f:
                self.assertEqual(b'b', f.read())

    def test_put_bytes(self):
        t = self.get_transport()

        if t.is_readonly():
            self.assertRaises(TransportNotPossible,
                              t.put_bytes, 'a', b'some text for a\n')
            return

        t.put_bytes('a', b'some text for a\n')
        self.assertTrue(t.has('a'))
        self.check_transport_contents(b'some text for a\n', t, 'a')

        # The contents should be overwritten
        t.put_bytes('a', b'new text for a\n')
        self.check_transport_contents(b'new text for a\n', t, 'a')

        self.assertRaises(NoSuchFile,
                          t.put_bytes, 'path/doesnt/exist/c', b'contents')

    def test_put_bytes_non_atomic(self):
        t = self.get_transport()

        if t.is_readonly():
            self.assertRaises(TransportNotPossible,
                              t.put_bytes_non_atomic, 'a', b'some text for a\n')
            return

        self.assertFalse(t.has('a'))
        t.put_bytes_non_atomic('a', b'some text for a\n')
        self.assertTrue(t.has('a'))
        self.check_transport_contents(b'some text for a\n', t, 'a')
        # Put also replaces contents
        t.put_bytes_non_atomic('a', b'new\ncontents for\na\n')
        self.check_transport_contents(b'new\ncontents for\na\n', t, 'a')

        # Make sure we can create another file
        t.put_bytes_non_atomic('d', b'contents for\nd\n')
        # And overwrite 'a' with empty contents
        t.put_bytes_non_atomic('a', b'')
        self.check_transport_contents(b'contents for\nd\n', t, 'd')
        self.check_transport_contents(b'', t, 'a')

        self.assertRaises(NoSuchFile, t.put_bytes_non_atomic, 'no/such/path',
                          b'contents\n')
        # Now test the create_parent flag
        self.assertRaises(NoSuchFile, t.put_bytes_non_atomic, 'dir/a',
                          b'contents\n')
        self.assertFalse(t.has('dir/a'))
        t.put_bytes_non_atomic('dir/a', b'contents for dir/a\n',
                               create_parent_dir=True)
        self.check_transport_contents(b'contents for dir/a\n', t, 'dir/a')

        # But we still get NoSuchFile if we can't make the parent dir
        self.assertRaises(NoSuchFile, t.put_bytes_non_atomic, 'not/there/a',
                          b'contents\n',
                          create_parent_dir=True)

    def test_put_bytes_permissions(self):
        t = self.get_transport()

        if t.is_readonly():
            return
        if not t._can_roundtrip_unix_modebits():
            # Can't roundtrip, so no need to run this test
            return
        t.put_bytes('mode644', b'test text\n', mode=0o644)
        self.assertTransportMode(t, 'mode644', 0o644)
        t.put_bytes('mode666', b'test text\n', mode=0o666)
        self.assertTransportMode(t, 'mode666', 0o666)
        t.put_bytes('mode600', b'test text\n', mode=0o600)
        self.assertTransportMode(t, 'mode600', 0o600)
        # Yes, you can put_bytes a file such that it becomes readonly
        t.put_bytes('mode400', b'test text\n', mode=0o400)
        self.assertTransportMode(t, 'mode400', 0o400)

        # The default permissions should be based on the current umask
        umask = osutils.get_umask()
        t.put_bytes('nomode', b'test text\n', mode=None)
        self.assertTransportMode(t, 'nomode', 0o666 & ~umask)

    def test_put_bytes_non_atomic_permissions(self):
        t = self.get_transport()

        if t.is_readonly():
            return
        if not t._can_roundtrip_unix_modebits():
            # Can't roundtrip, so no need to run this test
            return
        t.put_bytes_non_atomic('mode644', b'test text\n', mode=0o644)
        self.assertTransportMode(t, 'mode644', 0o644)
        t.put_bytes_non_atomic('mode666', b'test text\n', mode=0o666)
        self.assertTransportMode(t, 'mode666', 0o666)
        t.put_bytes_non_atomic('mode600', b'test text\n', mode=0o600)
        self.assertTransportMode(t, 'mode600', 0o600)
        t.put_bytes_non_atomic('mode400', b'test text\n', mode=0o400)
        self.assertTransportMode(t, 'mode400', 0o400)

        # The default permissions should be based on the current umask
        umask = osutils.get_umask()
        t.put_bytes_non_atomic('nomode', b'test text\n', mode=None)
        self.assertTransportMode(t, 'nomode', 0o666 & ~umask)

        # We should also be able to set the mode for a parent directory
        # when it is created
        t.put_bytes_non_atomic('dir700/mode664', b'test text\n', mode=0o664,
                               dir_mode=0o700, create_parent_dir=True)
        self.assertTransportMode(t, 'dir700', 0o700)
        t.put_bytes_non_atomic('dir770/mode664', b'test text\n', mode=0o664,
                               dir_mode=0o770, create_parent_dir=True)
        self.assertTransportMode(t, 'dir770', 0o770)
        t.put_bytes_non_atomic('dir777/mode664', b'test text\n', mode=0o664,
                               dir_mode=0o777, create_parent_dir=True)
        self.assertTransportMode(t, 'dir777', 0o777)

    def test_put_file(self):
        t = self.get_transport()

        if t.is_readonly():
            self.assertRaises(TransportNotPossible,
                              t.put_file, 'a', BytesIO(b'some text for a\n'))
            return

        result = t.put_file('a', BytesIO(b'some text for a\n'))
        # put_file returns the length of the data written
        self.assertEqual(16, result)
        self.assertTrue(t.has('a'))
        self.check_transport_contents(b'some text for a\n', t, 'a')
        # Put also replaces contents
        result = t.put_file('a', BytesIO(b'new\ncontents for\na\n'))
        self.assertEqual(19, result)
        self.check_transport_contents(b'new\ncontents for\na\n', t, 'a')
        self.assertRaises(NoSuchFile,
                          t.put_file, 'path/doesnt/exist/c',
                          BytesIO(b'contents'))

    def test_put_file_non_atomic(self):
        t = self.get_transport()

        if t.is_readonly():
            self.assertRaises(TransportNotPossible,
                              t.put_file_non_atomic, 'a', BytesIO(b'some text for a\n'))
            return

        self.assertFalse(t.has('a'))
        t.put_file_non_atomic('a', BytesIO(b'some text for a\n'))
        self.assertTrue(t.has('a'))
        self.check_transport_contents(b'some text for a\n', t, 'a')
        # Put also replaces contents
        t.put_file_non_atomic('a', BytesIO(b'new\ncontents for\na\n'))
        self.check_transport_contents(b'new\ncontents for\na\n', t, 'a')

        # Make sure we can create another file
        t.put_file_non_atomic('d', BytesIO(b'contents for\nd\n'))
        # And overwrite 'a' with empty contents
        t.put_file_non_atomic('a', BytesIO(b''))
        self.check_transport_contents(b'contents for\nd\n', t, 'd')
        self.check_transport_contents(b'', t, 'a')

        self.assertRaises(NoSuchFile, t.put_file_non_atomic, 'no/such/path',
                          BytesIO(b'contents\n'))
        # Now test the create_parent flag
        self.assertRaises(NoSuchFile, t.put_file_non_atomic, 'dir/a',
                          BytesIO(b'contents\n'))
        self.assertFalse(t.has('dir/a'))
        t.put_file_non_atomic('dir/a', BytesIO(b'contents for dir/a\n'),
                              create_parent_dir=True)
        self.check_transport_contents(b'contents for dir/a\n', t, 'dir/a')

        # But we still get NoSuchFile if we can't make the parent dir
        self.assertRaises(NoSuchFile, t.put_file_non_atomic, 'not/there/a',
                          BytesIO(b'contents\n'),
                          create_parent_dir=True)

    def test_put_file_permissions(self):

        t = self.get_transport()

        if t.is_readonly():
            return
        if not t._can_roundtrip_unix_modebits():
            # Can't roundtrip, so no need to run this test
            return
        t.put_file('mode644', BytesIO(b'test text\n'), mode=0o644)
        self.assertTransportMode(t, 'mode644', 0o644)
        t.put_file('mode666', BytesIO(b'test text\n'), mode=0o666)
        self.assertTransportMode(t, 'mode666', 0o666)
        t.put_file('mode600', BytesIO(b'test text\n'), mode=0o600)
        self.assertTransportMode(t, 'mode600', 0o600)
        # Yes, you can put a file such that it becomes readonly
        t.put_file('mode400', BytesIO(b'test text\n'), mode=0o400)
        self.assertTransportMode(t, 'mode400', 0o400)
        # The default permissions should be based on the current umask
        umask = osutils.get_umask()
        t.put_file('nomode', BytesIO(b'test text\n'), mode=None)
        self.assertTransportMode(t, 'nomode', 0o666 & ~umask)

    def test_put_file_non_atomic_permissions(self):
        t = self.get_transport()

        if t.is_readonly():
            return
        if not t._can_roundtrip_unix_modebits():
            # Can't roundtrip, so no need to run this test
            return
        t.put_file_non_atomic('mode644', BytesIO(b'test text\n'), mode=0o644)
        self.assertTransportMode(t, 'mode644', 0o644)
        t.put_file_non_atomic('mode666', BytesIO(b'test text\n'), mode=0o666)
        self.assertTransportMode(t, 'mode666', 0o666)
        t.put_file_non_atomic('mode600', BytesIO(b'test text\n'), mode=0o600)
        self.assertTransportMode(t, 'mode600', 0o600)
        # Yes, you can put_file_non_atomic a file such that it becomes readonly
        t.put_file_non_atomic('mode400', BytesIO(b'test text\n'), mode=0o400)
        self.assertTransportMode(t, 'mode400', 0o400)

        # The default permissions should be based on the current umask
        umask = osutils.get_umask()
        t.put_file_non_atomic('nomode', BytesIO(b'test text\n'), mode=None)
        self.assertTransportMode(t, 'nomode', 0o666 & ~umask)

        # We should also be able to set the mode for a parent directory
        # when it is created
        sio = BytesIO()
        t.put_file_non_atomic('dir700/mode664', sio, mode=0o664,
                              dir_mode=0o700, create_parent_dir=True)
        self.assertTransportMode(t, 'dir700', 0o700)
        t.put_file_non_atomic('dir770/mode664', sio, mode=0o664,
                              dir_mode=0o770, create_parent_dir=True)
        self.assertTransportMode(t, 'dir770', 0o770)
        t.put_file_non_atomic('dir777/mode664', sio, mode=0o664,
                              dir_mode=0o777, create_parent_dir=True)
        self.assertTransportMode(t, 'dir777', 0o777)

    def test_put_bytes_unicode(self):
        t = self.get_transport()
        if t.is_readonly():
            return
        unicode_string = '\u1234'
        self.assertRaises(TypeError, t.put_bytes, 'foo', unicode_string)

    def test_mkdir(self):
        t = self.get_transport()

        if t.is_readonly():
            # cannot mkdir on readonly transports. We're not testing for
            # cache coherency because cache behaviour is not currently
            # defined for the transport interface.
            self.assertRaises(TransportNotPossible, t.mkdir, '.')
            self.assertRaises(TransportNotPossible, t.mkdir, 'new_dir')
            self.assertRaises(TransportNotPossible,
                              t.mkdir, 'path/doesnt/exist')
            return
        # Test mkdir
        t.mkdir('dir_a')
        self.assertEqual(t.has('dir_a'), True)
        self.assertEqual(t.has('dir_b'), False)

        t.mkdir('dir_b')
        self.assertEqual(t.has('dir_b'), True)

        self.assertEqual([t.has(n) for n in
                          ['dir_a', 'dir_b', 'dir_q', 'dir_b']],
                         [True, True, False, True])

        # we were testing that a local mkdir followed by a transport
        # mkdir failed thusly, but given that we * in one process * do not
        # concurrently fiddle with disk dirs and then use transport to do
        # things, the win here seems marginal compared to the constraint on
        # the interface. RBC 20051227
        t.mkdir('dir_g')
        self.assertRaises(FileExists, t.mkdir, 'dir_g')

        # Test get/put in sub-directories
        t.put_bytes('dir_a/a', b'contents of dir_a/a')
        t.put_file('dir_b/b', BytesIO(b'contents of dir_b/b'))
        self.check_transport_contents(b'contents of dir_a/a', t, 'dir_a/a')
        self.check_transport_contents(b'contents of dir_b/b', t, 'dir_b/b')

        # mkdir of a dir with an absent parent
        self.assertRaises(NoSuchFile, t.mkdir, 'missing/dir')

    def test_mkdir_permissions(self):
        t = self.get_transport()
        if t.is_readonly():
            return
        if not t._can_roundtrip_unix_modebits():
            # no sense testing on this transport
            return
        # Test mkdir with a mode
        t.mkdir('dmode755', mode=0o755)
        self.assertTransportMode(t, 'dmode755', 0o755)
        t.mkdir('dmode555', mode=0o555)
        self.assertTransportMode(t, 'dmode555', 0o555)
        t.mkdir('dmode777', mode=0o777)
        self.assertTransportMode(t, 'dmode777', 0o777)
        t.mkdir('dmode700', mode=0o700)
        self.assertTransportMode(t, 'dmode700', 0o700)
        t.mkdir('mdmode755', mode=0o755)
        self.assertTransportMode(t, 'mdmode755', 0o755)

        # Default mode should be based on umask
        umask = osutils.get_umask()
        t.mkdir('dnomode', mode=None)
        self.assertTransportMode(t, 'dnomode', 0o777 & ~umask)

    def test_opening_a_file_stream_creates_file(self):
        t = self.get_transport()
        if t.is_readonly():
            return
        handle = t.open_write_stream('foo')
        try:
            self.assertEqual(b'', t.get_bytes('foo'))
        finally:
            handle.close()

    def test_opening_a_file_stream_can_set_mode(self):
        t = self.get_transport()
        if t.is_readonly():
            self.assertRaises((TransportNotPossible, NotImplementedError),
                              t.open_write_stream, 'foo')
            return
        if not t._can_roundtrip_unix_modebits():
            # Can't roundtrip, so no need to run this test
            return

        def check_mode(name, mode, expected):
            handle = t.open_write_stream(name, mode=mode)
            handle.close()
            self.assertTransportMode(t, name, expected)
        check_mode('mode644', 0o644, 0o644)
        check_mode('mode666', 0o666, 0o666)
        check_mode('mode600', 0o600, 0o600)
        # The default permissions should be based on the current umask
        check_mode('nomode', None, 0o666 & ~osutils.get_umask())

    def test_copy_to(self):
        # FIXME: test:   same server to same server (partly done)
        # same protocol two servers
        # and    different protocols (done for now except for MemoryTransport.
        # - RBC 20060122

        def simple_copy_files(transport_from, transport_to):
            files = ['a', 'b', 'c', 'd']
            self.build_tree(files, transport=transport_from)
            self.assertEqual(4, transport_from.copy_to(files, transport_to))
            for f in files:
                self.check_transport_contents(transport_to.get_bytes(f),
                                              transport_from, f)

        t = self.get_transport()
        if t.__class__.__name__ == "SFTPTransport":
            self.skipTest("SFTP copy_to currently too flakey to use")
        temp_transport = MemoryTransport('memory:///')
        simple_copy_files(t, temp_transport)
        if not t.is_readonly():
            t.mkdir('copy_to_simple')
            t2 = t.clone('copy_to_simple')
            simple_copy_files(t, t2)

        # Test that copying into a missing directory raises
        # NoSuchFile
        if t.is_readonly():
            self.build_tree(['e/', 'e/f'])
        else:
            t.mkdir('e')
            t.put_bytes('e/f', b'contents of e')
        self.assertRaises(NoSuchFile, t.copy_to, ['e/f'], temp_transport)
        temp_transport.mkdir('e')
        t.copy_to(['e/f'], temp_transport)

        del temp_transport
        temp_transport = MemoryTransport('memory:///')

        files = ['a', 'b', 'c', 'd']
        t.copy_to(iter(files), temp_transport)
        for f in files:
            self.check_transport_contents(temp_transport.get_bytes(f),
                                          t, f)
        del temp_transport

        for mode in (0o666, 0o644, 0o600, 0o400):
            temp_transport = MemoryTransport("memory:///")
            t.copy_to(files, temp_transport, mode=mode)
            for f in files:
                self.assertTransportMode(temp_transport, f, mode)

    def test_create_prefix(self):
        t = self.get_transport()
        sub = t.clone('foo').clone('bar')
        try:
            sub.create_prefix()
        except TransportNotPossible:
            self.assertTrue(t.is_readonly())
        else:
            self.assertTrue(t.has('foo/bar'))

    def test_append_file(self):
        t = self.get_transport()

        if t.is_readonly():
            self.assertRaises(TransportNotPossible,
                              t.append_file, 'a', 'add\nsome\nmore\ncontents\n')
            return
        t.put_bytes('a', b'diff\ncontents for\na\n')
        t.put_bytes('b', b'contents\nfor b\n')

        self.assertEqual(20,
                         t.append_file('a', BytesIO(b'add\nsome\nmore\ncontents\n')))

        self.check_transport_contents(
            b'diff\ncontents for\na\nadd\nsome\nmore\ncontents\n',
            t, 'a')

        # a file with no parent should fail..
        self.assertRaises(NoSuchFile,
                          t.append_file, 'missing/path', BytesIO(b'content'))

        # And we can create new files, too
        self.assertEqual(0,
                         t.append_file('c', BytesIO(b'some text\nfor a missing file\n')))
        self.check_transport_contents(b'some text\nfor a missing file\n',
                                      t, 'c')

    def test_append_bytes(self):
        t = self.get_transport()

        if t.is_readonly():
            self.assertRaises(TransportNotPossible,
                              t.append_bytes, 'a', b'add\nsome\nmore\ncontents\n')
            return

        self.assertEqual(0, t.append_bytes('a', b'diff\ncontents for\na\n'))
        self.assertEqual(0, t.append_bytes('b', b'contents\nfor b\n'))

        self.assertEqual(20,
                         t.append_bytes('a', b'add\nsome\nmore\ncontents\n'))

        self.check_transport_contents(
            b'diff\ncontents for\na\nadd\nsome\nmore\ncontents\n',
            t, 'a')

        # a file with no parent should fail..
        self.assertRaises(NoSuchFile,
                          t.append_bytes, 'missing/path', b'content')

    def test_append_file_mode(self):
        """Check that append accepts a mode parameter"""
        # check append accepts a mode
        t = self.get_transport()
        if t.is_readonly():
            self.assertRaises(TransportNotPossible,
                              t.append_file, 'f', BytesIO(b'f'), mode=None)
            return
        t.append_file('f', BytesIO(b'f'), mode=None)

    def test_append_bytes_mode(self):
        # check append_bytes accepts a mode
        t = self.get_transport()
        if t.is_readonly():
            self.assertRaises(TransportNotPossible,
                              t.append_bytes, 'f', b'f', mode=None)
            return
        t.append_bytes('f', b'f', mode=None)

    def test_delete(self):
        # TODO: Test Transport.delete
        t = self.get_transport()

        # Not much to do with a readonly transport
        if t.is_readonly():
            self.assertRaises(TransportNotPossible, t.delete, 'missing')
            return

        t.put_bytes('a', b'a little bit of text\n')
        self.assertTrue(t.has('a'))
        t.delete('a')
        self.assertFalse(t.has('a'))

        self.assertRaises(NoSuchFile, t.delete, 'a')

        t.put_bytes('a', b'a text\n')
        t.put_bytes('b', b'b text\n')
        t.put_bytes('c', b'c text\n')
        self.assertEqual([True, True, True],
                         [t.has(n) for n in ['a', 'b', 'c']])
        t.delete('a')
        t.delete('c')
        self.assertEqual([False, True, False],
                         [t.has(n) for n in ['a', 'b', 'c']])
        self.assertFalse(t.has('a'))
        self.assertTrue(t.has('b'))
        self.assertFalse(t.has('c'))

        for name in ['a', 'c', 'd']:
            self.assertRaises(NoSuchFile, t.delete, name)

        # We should have deleted everything
        # SftpServer creates control files in the
        # working directory, so we can just do a
        # plain "listdir".
        # self.assertEqual([], os.listdir('.'))

    def test_recommended_page_size(self):
        """Transports recommend a page size for partial access to files."""
        t = self.get_transport()
        self.assertIsInstance(t.recommended_page_size(), int)

    def test_rmdir(self):
        t = self.get_transport()
        # Not much to do with a readonly transport
        if t.is_readonly():
            self.assertRaises(TransportNotPossible, t.rmdir, 'missing')
            return
        t.mkdir('adir')
        t.mkdir('adir/bdir')
        t.rmdir('adir/bdir')
        # ftp may not be able to raise NoSuchFile for lack of
        # details when failing
        self.assertRaises((NoSuchFile, PathError), t.rmdir, 'adir/bdir')
        t.rmdir('adir')
        self.assertRaises((NoSuchFile, PathError), t.rmdir, 'adir')

    def test_rmdir_not_empty(self):
        """Deleting a non-empty directory raises an exception

        sftp (and possibly others) don't give us a specific "directory not
        empty" exception -- we can just see that the operation failed.
        """
        t = self.get_transport()
        if t.is_readonly():
            return
        t.mkdir('adir')
        t.mkdir('adir/bdir')
        self.assertRaises(PathError, t.rmdir, 'adir')

    def test_rmdir_empty_but_similar_prefix(self):
        """rmdir does not get confused by sibling paths.

        A naive implementation of MemoryTransport would refuse to rmdir
        ".bzr/branch" if there is a ".bzr/branch-format" directory, because it
        uses "path.startswith(dir)" on all file paths to determine if directory
        is empty.
        """
        t = self.get_transport()
        if t.is_readonly():
            return
        t.mkdir('foo')
        t.put_bytes('foo-bar', b'')
        t.mkdir('foo-baz')
        t.rmdir('foo')
        self.assertRaises((NoSuchFile, PathError), t.rmdir, 'foo')
        self.assertTrue(t.has('foo-bar'))

    def test_rename_dir_succeeds(self):
        t = self.get_transport()
        if t.is_readonly():
            self.assertRaises((TransportNotPossible, NotImplementedError),
                              t.rename, 'foo', 'bar')
            return
        t.mkdir('adir')
        t.mkdir('adir/asubdir')
        t.rename('adir', 'bdir')
        self.assertTrue(t.has('bdir/asubdir'))
        self.assertFalse(t.has('adir'))

    def test_rename_dir_nonempty(self):
        """Attempting to replace a nonemtpy directory should fail"""
        t = self.get_transport()
        if t.is_readonly():
            self.assertRaises((TransportNotPossible, NotImplementedError),
                              t.rename, 'foo', 'bar')
            return
        t.mkdir('adir')
        t.mkdir('adir/asubdir')
        t.mkdir('bdir')
        t.mkdir('bdir/bsubdir')
        # any kind of PathError would be OK, though we normally expect
        # DirectoryNotEmpty
        self.assertRaises(PathError, t.rename, 'bdir', 'adir')
        # nothing was changed so it should still be as before
        self.assertTrue(t.has('bdir/bsubdir'))
        self.assertFalse(t.has('adir/bdir'))
        self.assertFalse(t.has('adir/bsubdir'))

    def test_rename_across_subdirs(self):
        t = self.get_transport()
        if t.is_readonly():
            raise TestNotApplicable("transport is readonly")
        t.mkdir('a')
        t.mkdir('b')
        ta = t.clone('a')
        tb = t.clone('b')
        ta.put_bytes('f', b'aoeu')
        ta.rename('f', '../b/f')
        self.assertTrue(tb.has('f'))
        self.assertFalse(ta.has('f'))
        self.assertTrue(t.has('b/f'))

    def test_delete_tree(self):
        t = self.get_transport()

        # Not much to do with a readonly transport
        if t.is_readonly():
            self.assertRaises(TransportNotPossible, t.delete_tree, 'missing')
            return

        # and does it like listing ?
        t.mkdir('adir')
        try:
            t.delete_tree('adir')
        except TransportNotPossible:
            # ok, this transport does not support delete_tree
            return

        # did it delete that trivial case?
        self.assertRaises(NoSuchFile, t.stat, 'adir')

        self.build_tree(['adir/',
                         'adir/file',
                         'adir/subdir/',
                         'adir/subdir/file',
                         'adir/subdir2/',
                         'adir/subdir2/file',
                         ], transport=t)

        t.delete_tree('adir')
        # adir should be gone now.
        self.assertRaises(NoSuchFile, t.stat, 'adir')

    def test_move(self):
        t = self.get_transport()

        if t.is_readonly():
            return

        # TODO: I would like to use os.listdir() to
        # make sure there are no extra files, but SftpServer
        # creates control files in the working directory
        # perhaps all of this could be done in a subdirectory

        t.put_bytes('a', b'a first file\n')
        self.assertEqual([True, False], [t.has(n) for n in ['a', 'b']])

        t.move('a', 'b')
        self.assertTrue(t.has('b'))
        self.assertFalse(t.has('a'))

        self.check_transport_contents(b'a first file\n', t, 'b')
        self.assertEqual([False, True], [t.has(n) for n in ['a', 'b']])

        # Overwrite a file
        t.put_bytes('c', b'c this file\n')
        t.move('c', 'b')
        self.assertFalse(t.has('c'))
        self.check_transport_contents(b'c this file\n', t, 'b')

        # TODO: Try to write a test for atomicity
        # TODO: Test moving into a non-existent subdirectory

    def test_copy(self):
        t = self.get_transport()

        if t.is_readonly():
            return

        t.put_bytes('a', b'a file\n')
        t.copy('a', 'b')
        self.check_transport_contents(b'a file\n', t, 'b')

        self.assertRaises(NoSuchFile, t.copy, 'c', 'd')
        os.mkdir('c')
        # What should the assert be if you try to copy a
        # file over a directory?
        #self.assertRaises(Something, t.copy, 'a', 'c')
        t.put_bytes('d', b'text in d\n')
        t.copy('d', 'b')
        self.check_transport_contents(b'text in d\n', t, 'b')

    def test_connection_error(self):
        """ConnectionError is raised when connection is impossible.

        The error should be raised from the first operation on the transport.
        """
        try:
            url = self._server.get_bogus_url()
        except NotImplementedError:
            raise TestSkipped("Transport %s has no bogus URL support." %
                              self._server.__class__)
        t = _mod_transport.get_transport_from_url(url)
        self.assertRaises((ConnectionError, NoSuchFile), t.get, '.bzr/branch')

    def test_stat(self):
        # TODO: Test stat, just try once, and if it throws, stop testing
        from stat import S_ISDIR, S_ISREG

        t = self.get_transport()

        try:
            st = t.stat('.')
        except TransportNotPossible as e:
            # This transport cannot stat
            return

        paths = ['a', 'b/', 'b/c', 'b/d/', 'b/d/e']
        sizes = [14, 0, 16, 0, 18]
        self.build_tree(paths, transport=t, line_endings='binary')

        for path, size in zip(paths, sizes):
            st = t.stat(path)
            if path.endswith('/'):
                self.assertTrue(S_ISDIR(st.st_mode))
                # directory sizes are meaningless
            else:
                self.assertTrue(S_ISREG(st.st_mode))
                self.assertEqual(size, st.st_size)

        self.assertRaises(NoSuchFile, t.stat, 'q')
        self.assertRaises(NoSuchFile, t.stat, 'b/a')

        self.build_tree(['subdir/', 'subdir/file'], transport=t)
        subdir = t.clone('subdir')
        st = subdir.stat('./file')
        st = subdir.stat('.')

    def test_hardlink(self):
        from stat import ST_NLINK

        t = self.get_transport()

        source_name = "original_target"
        link_name = "target_link"

        self.build_tree([source_name], transport=t)

        try:
            t.hardlink(source_name, link_name)

            self.assertTrue(t.has(source_name))
            self.assertTrue(t.has(link_name))

            st = t.stat(link_name)
            self.assertEqual(st[ST_NLINK], 2)
        except TransportNotPossible:
            raise TestSkipped("Transport %s does not support hardlinks." %
                              self._server.__class__)

    def test_symlink(self):
        from stat import S_ISLNK

        t = self.get_transport()

        source_name = "original_target"
        link_name = "target_link"

        self.build_tree([source_name], transport=t)

        try:
            t.symlink(source_name, link_name)

            self.assertTrue(t.has(source_name))
            self.assertTrue(t.has(link_name))

            st = t.stat(link_name)
            self.assertTrue(S_ISLNK(st.st_mode),
                            "expected symlink, got mode %o" % st.st_mode)
        except TransportNotPossible:
            raise TestSkipped("Transport %s does not support symlinks." %
                              self._server.__class__)

        self.assertEqual(source_name, t.readlink(link_name))

    def test_readlink_nonexistent(self):
        t = self.get_transport()
        try:
            self.assertRaises(NoSuchFile, t.readlink, 'nonexistent')
        except TransportNotPossible:
            raise TestSkipped("Transport %s does not support symlinks." %
                              self._server.__class__)

    def test_list_dir(self):
        # TODO: Test list_dir, just try once, and if it throws, stop testing
        t = self.get_transport()

        if not t.listable():
            self.assertRaises(TransportNotPossible, t.list_dir, '.')
            return

        def sorted_list(d, transport):
            l = sorted(transport.list_dir(d))
            return l

        self.assertEqual([], sorted_list('.', t))
        # c2 is precisely one letter longer than c here to test that
        # suffixing is not confused.
        # a%25b checks that quoting is done consistently across transports
        tree_names = ['a', 'a%25b', 'b', 'c/', 'c/d', 'c/e', 'c2/']

        if not t.is_readonly():
            self.build_tree(tree_names, transport=t)
        else:
            self.build_tree(tree_names)

        self.assertEqual(
            ['a', 'a%2525b', 'b', 'c', 'c2'], sorted_list('', t))
        self.assertEqual(
            ['a', 'a%2525b', 'b', 'c', 'c2'], sorted_list('.', t))
        self.assertEqual(['d', 'e'], sorted_list('c', t))

        # Cloning the transport produces an equivalent listing
        self.assertEqual(['d', 'e'], sorted_list('', t.clone('c')))

        if not t.is_readonly():
            t.delete('c/d')
            t.delete('b')
        else:
            os.unlink('c/d')
            os.unlink('b')

        self.assertEqual(['a', 'a%2525b', 'c', 'c2'], sorted_list('.', t))
        self.assertEqual(['e'], sorted_list('c', t))

        self.assertListRaises(PathError, t.list_dir, 'q')
        self.assertListRaises(PathError, t.list_dir, 'c/f')
        # 'a' is a file, list_dir should raise an error
        self.assertListRaises(PathError, t.list_dir, 'a')

    def test_list_dir_result_is_url_escaped(self):
        t = self.get_transport()
        if not t.listable():
            raise TestSkipped("transport not listable")

        if not t.is_readonly():
            self.build_tree(['a/', 'a/%'], transport=t)
        else:
            self.build_tree(['a/', 'a/%'])

        names = list(t.list_dir('a'))
        self.assertEqual(['%25'], names)
        self.assertIsInstance(names[0], str)

    def test_clone_preserve_info(self):
        t1 = self.get_transport()
        if not isinstance(t1, ConnectedTransport):
            raise TestSkipped("not a connected transport")

        t2 = t1.clone('subdir')
        self.assertEqual(t1._parsed_url.scheme, t2._parsed_url.scheme)
        self.assertEqual(t1._parsed_url.user, t2._parsed_url.user)
        self.assertEqual(t1._parsed_url.password, t2._parsed_url.password)
        self.assertEqual(t1._parsed_url.host, t2._parsed_url.host)
        self.assertEqual(t1._parsed_url.port, t2._parsed_url.port)

    def test__reuse_for(self):
        t = self.get_transport()
        if not isinstance(t, ConnectedTransport):
            raise TestSkipped("not a connected transport")

        def new_url(scheme=None, user=None, password=None,
                    host=None, port=None, path=None):
            """Build a new url from t.base changing only parts of it.

            Only the parameters different from None will be changed.
            """
            if scheme is None:
                scheme = t._parsed_url.scheme
            if user is None:
                user = t._parsed_url.user
            if password is None:
                password = t._parsed_url.password
            if user is None:
                user = t._parsed_url.user
            if host is None:
                host = t._parsed_url.host
            if port is None:
                port = t._parsed_url.port
            if path is None:
                path = t._parsed_url.path
            return str(urlutils.URL(scheme, user, password, host, port, path))

        if t._parsed_url.scheme == 'ftp':
            scheme = 'sftp'
        else:
            scheme = 'ftp'
        self.assertIsNot(t, t._reuse_for(new_url(scheme=scheme)))
        if t._parsed_url.user == 'me':
            user = 'you'
        else:
            user = 'me'
        self.assertIsNot(t, t._reuse_for(new_url(user=user)))
        # passwords are not taken into account because:
        # - it makes no sense to have two different valid passwords for the
        #   same user
        # - _password in ConnectedTransport is intended to collect what the
        #   user specified from the command-line and there are cases where the
        #   new url can contain no password (if the url was built from an
        #   existing transport.base for example)
        # - password are considered part of the credentials provided at
        #   connection creation time and as such may not be present in the url
        #   (they may be typed by the user when prompted for example)
        self.assertIs(t, t._reuse_for(new_url(password='from space')))
        # We will not connect, we can use a invalid host
        self.assertIsNot(t, t._reuse_for(
            new_url(host=t._parsed_url.host + 'bar')))
        if t._parsed_url.port == 1234:
            port = 4321
        else:
            port = 1234
        self.assertIsNot(t, t._reuse_for(new_url(port=port)))
        # No point in trying to reuse a transport for a local URL
        self.assertIs(None, t._reuse_for('/valid_but_not_existing'))

    def test_connection_sharing(self):
        t = self.get_transport()
        if not isinstance(t, ConnectedTransport):
            raise TestSkipped("not a connected transport")

        c = t.clone('subdir')
        # Some transports will create the connection  only when needed
        t.has('surely_not')  # Force connection
        self.assertIs(t._get_connection(), c._get_connection())

        # Temporary failure, we need to create a new dummy connection
        new_connection = None
        t._set_connection(new_connection)
        # Check that both transports use the same connection
        self.assertIs(new_connection, t._get_connection())
        self.assertIs(new_connection, c._get_connection())

    def test_reuse_connection_for_various_paths(self):
        t = self.get_transport()
        if not isinstance(t, ConnectedTransport):
            raise TestSkipped("not a connected transport")

        t.has('surely_not')  # Force connection
        self.assertIsNot(None, t._get_connection())

        subdir = t._reuse_for(t.base + 'whatever/but/deep/down/the/path')
        self.assertIsNot(t, subdir)
        self.assertIs(t._get_connection(), subdir._get_connection())

        home = subdir._reuse_for(t.base + 'home')
        self.assertIs(t._get_connection(), home._get_connection())
        self.assertIs(subdir._get_connection(), home._get_connection())

    def test_clone(self):
        # TODO: Test that clone moves up and down the filesystem
        t1 = self.get_transport()

        self.build_tree(['a', 'b/', 'b/c'], transport=t1)

        self.assertTrue(t1.has('a'))
        self.assertTrue(t1.has('b/c'))
        self.assertFalse(t1.has('c'))

        t2 = t1.clone('b')
        self.assertEqual(t1.base + 'b/', t2.base)

        self.assertTrue(t2.has('c'))
        self.assertFalse(t2.has('a'))

        t3 = t2.clone('..')
        self.assertTrue(t3.has('a'))
        self.assertFalse(t3.has('c'))

        self.assertFalse(t1.has('b/d'))
        self.assertFalse(t2.has('d'))
        self.assertFalse(t3.has('b/d'))

        if t1.is_readonly():
            self.build_tree_contents([('b/d', b'newfile\n')])
        else:
            t2.put_bytes('d', b'newfile\n')

        self.assertTrue(t1.has('b/d'))
        self.assertTrue(t2.has('d'))
        self.assertTrue(t3.has('b/d'))

    def test_clone_to_root(self):
        orig_transport = self.get_transport()
        # Repeatedly go up to a parent directory until we're at the root
        # directory of this transport
        root_transport = orig_transport
        new_transport = root_transport.clone("..")
        # as we are walking up directories, the path must be
        # growing less, except at the top
        self.assertTrue(len(new_transport.base) < len(root_transport.base) or
                        new_transport.base == root_transport.base)
        while new_transport.base != root_transport.base:
            root_transport = new_transport
            new_transport = root_transport.clone("..")
            # as we are walking up directories, the path must be
            # growing less, except at the top
            self.assertTrue(len(new_transport.base) < len(root_transport.base) or
                            new_transport.base == root_transport.base)

        # Cloning to "/" should take us to exactly the same location.
        self.assertEqual(root_transport.base, orig_transport.clone("/").base)
        # the abspath of "/" from the original transport should be the same
        # as the base at the root:
        self.assertEqual(orig_transport.abspath("/"), root_transport.base)

        # At the root, the URL must still end with / as its a directory
        self.assertEqual(root_transport.base[-1], '/')

    def test_clone_from_root(self):
        """At the root, cloning to a simple dir should just do string append."""
        orig_transport = self.get_transport()
        root_transport = orig_transport.clone('/')
        self.assertEqual(root_transport.base + '.bzr/',
                         root_transport.clone('.bzr').base)

    def test_base_url(self):
        t = self.get_transport()
        self.assertEqual('/', t.base[-1])

    def test_relpath(self):
        t = self.get_transport()
        self.assertEqual('', t.relpath(t.base))
        # base ends with /
        self.assertEqual('', t.relpath(t.base[:-1]))
        # subdirs which don't exist should still give relpaths.
        self.assertEqual('foo', t.relpath(t.base + 'foo'))
        # trailing slash should be the same.
        self.assertEqual('foo', t.relpath(t.base + 'foo/'))

    def test_relpath_at_root(self):
        t = self.get_transport()
        # clone all the way to the top
        new_transport = t.clone('..')
        while new_transport.base != t.base:
            t = new_transport
            new_transport = t.clone('..')
        # we must be able to get a relpath below the root
        self.assertEqual('', t.relpath(t.base))
        # and a deeper one should work too
        self.assertEqual('foo/bar', t.relpath(t.base + 'foo/bar'))

    def test_abspath(self):
        # smoke test for abspath. Corner cases for backends like unix fs's
        # that have aliasing problems like symlinks should go in backend
        # specific test cases.
        transport = self.get_transport()

        self.assertEqual(transport.base + 'relpath',
                         transport.abspath('relpath'))

        # This should work without raising an error.
        transport.abspath("/")

        # the abspath of "/" and "/foo/.." should result in the same location
        self.assertEqual(transport.abspath("/"), transport.abspath("/foo/.."))

        self.assertEqual(transport.clone("/").abspath('foo'),
                         transport.abspath("/foo"))

    # GZ 2011-01-26: Test in per_transport but not using self.get_transport?
    def test_win32_abspath(self):
        # Note: we tried to set sys.platform='win32' so we could test on
        # other platforms too, but then osutils does platform specific
        # things at import time which defeated us...
        if sys.platform != 'win32':
            raise TestSkipped(
                'Testing drive letters in abspath implemented only for win32')

        # smoke test for abspath on win32.
        # a transport based on 'file:///' never fully qualifies the drive.
        transport = _mod_transport.get_transport_from_url("file:///")
        self.assertEqual(transport.abspath("/"), "file:///")

        # but a transport that starts with a drive spec must keep it.
        transport = _mod_transport.get_transport_from_url("file:///C:/")
        self.assertEqual(transport.abspath("/"), "file:///C:/")

    def test_local_abspath(self):
        transport = self.get_transport()
        try:
            p = transport.local_abspath('.')
        except (errors.NotLocalUrl, TransportNotPossible) as e:
            # should be formattable
            s = str(e)
        else:
            self.assertEqual(getcwd(), p)

    def test_abspath_at_root(self):
        t = self.get_transport()
        # clone all the way to the top
        new_transport = t.clone('..')
        while new_transport.base != t.base:
            t = new_transport
            new_transport = t.clone('..')
        # we must be able to get a abspath of the root when we ask for
        # t.abspath('..') - this due to our choice that clone('..')
        # should return the root from the root, combined with the desire that
        # the url from clone('..') and from abspath('..') should be the same.
        self.assertEqual(t.base, t.abspath('..'))
        # '' should give us the root
        self.assertEqual(t.base, t.abspath(''))
        # and a path should append to the url
        self.assertEqual(t.base + 'foo', t.abspath('foo'))

    def test_iter_files_recursive(self):
        transport = self.get_transport()
        if not transport.listable():
            self.assertRaises(TransportNotPossible,
                              transport.iter_files_recursive)
            return
        self.build_tree(['isolated/',
                         'isolated/dir/',
                         'isolated/dir/foo',
                         'isolated/dir/bar',
                         'isolated/dir/b%25z',  # make sure quoting is correct
                         'isolated/bar'],
                        transport=transport)
        paths = set(transport.iter_files_recursive())
        # nb the directories are not converted
        self.assertEqual(paths,
                         {'isolated/dir/foo',
                          'isolated/dir/bar',
                          'isolated/dir/b%2525z',
                          'isolated/bar'})
        sub_transport = transport.clone('isolated')
        paths = set(sub_transport.iter_files_recursive())
        self.assertEqual(paths,
                         {'dir/foo', 'dir/bar', 'dir/b%2525z', 'bar'})

    def test_copy_tree(self):
        # TODO: test file contents and permissions are preserved. This test was
        # added just to ensure that quoting was handled correctly.
        # -- David Allouche 2006-08-11
        transport = self.get_transport()
        if not transport.listable():
            self.assertRaises(TransportNotPossible,
                              transport.iter_files_recursive)
            return
        if transport.is_readonly():
            return
        self.build_tree(['from/',
                         'from/dir/',
                         'from/dir/foo',
                         'from/dir/bar',
                         'from/dir/b%25z',  # make sure quoting is correct
                         'from/bar'],
                        transport=transport)
        transport.copy_tree('from', 'to')
        paths = set(transport.iter_files_recursive())
        self.assertEqual(paths,
                         {'from/dir/foo',
                          'from/dir/bar',
                          'from/dir/b%2525z',
                          'from/bar',
                          'to/dir/foo',
                          'to/dir/bar',
                          'to/dir/b%2525z',
                          'to/bar', })

    def test_copy_tree_to_transport(self):
        transport = self.get_transport()
        if not transport.listable():
            self.assertRaises(TransportNotPossible,
                              transport.iter_files_recursive)
            return
        if transport.is_readonly():
            return
        self.build_tree(['from/',
                         'from/dir/',
                         'from/dir/foo',
                         'from/dir/bar',
                         'from/dir/b%25z',  # make sure quoting is correct
                         'from/bar'],
                        transport=transport)
        from_transport = transport.clone('from')
        to_transport = transport.clone('to')
        to_transport.ensure_base()
        from_transport.copy_tree_to_transport(to_transport)
        paths = set(transport.iter_files_recursive())
        self.assertEqual(paths,
                         {'from/dir/foo',
                          'from/dir/bar',
                          'from/dir/b%2525z',
                          'from/bar',
                          'to/dir/foo',
                          'to/dir/bar',
                          'to/dir/b%2525z',
                          'to/bar', })

    def test_unicode_paths(self):
        """Test that we can read/write files with Unicode names."""
        t = self.get_transport()

        # With FAT32 and certain encodings on win32
        # '\xe5' and '\xe4' actually map to the same file
        # adding a suffix kicks in the 'preserving but insensitive'
        # route, and maintains the right files
        files = ['\xe5.1',  # a w/ circle iso-8859-1
                 '\xe4.2',  # a w/ dots iso-8859-1
                 '\u017d',  # Z with umlat iso-8859-2
                 '\u062c',  # Arabic j
                 '\u0410',  # Russian A
                 '\u65e5',  # Kanji person
                 ]

        no_unicode_support = getattr(self._server, 'no_unicode_support', False)
        if no_unicode_support:
            self.knownFailure("test server cannot handle unicode paths")

        try:
            self.build_tree(files, transport=t, line_endings='binary')
        except UnicodeError:
            raise TestSkipped(
                "cannot handle unicode paths in current encoding")

        # A plain unicode string is not a valid url
        for fname in files:
            self.assertRaises(urlutils.InvalidURL, t.get, fname)

        for fname in files:
            fname_utf8 = fname.encode('utf-8')
            contents = b'contents of %s\n' % (fname_utf8,)
            self.check_transport_contents(contents, t, urlutils.escape(fname))

    def test_connect_twice_is_same_content(self):
        # check that our server (whatever it is) is accessible reliably
        # via get_transport and multiple connections share content.
        transport = self.get_transport()
        if transport.is_readonly():
            return
        transport.put_bytes('foo', b'bar')
        transport3 = self.get_transport()
        self.check_transport_contents(b'bar', transport3, 'foo')

        # now opening at a relative url should give use a sane result:
        transport.mkdir('newdir')
        transport5 = self.get_transport('newdir')
        transport6 = transport5.clone('..')
        self.check_transport_contents(b'bar', transport6, 'foo')

    def test_lock_write(self):
        """Test transport-level write locks.

        These are deprecated and transports may decline to support them.
        """
        transport = self.get_transport()
        if transport.is_readonly():
            self.assertRaises(TransportNotPossible,
                              transport.lock_write, 'foo')
            return
        transport.put_bytes('lock', b'')
        try:
            lock = transport.lock_write('lock')
        except TransportNotPossible:
            return
        # TODO make this consistent on all platforms:
        # self.assertRaises(LockError, transport.lock_write, 'lock')
        lock.unlock()

    def test_lock_read(self):
        """Test transport-level read locks.

        These are deprecated and transports may decline to support them.
        """
        transport = self.get_transport()
        if transport.is_readonly():
            open('lock', 'w').close()
        else:
            transport.put_bytes('lock', b'')
        try:
            lock = transport.lock_read('lock')
        except TransportNotPossible:
            return
        # TODO make this consistent on all platforms:
        # self.assertRaises(LockError, transport.lock_read, 'lock')
        lock.unlock()

    def test_readv(self):
        transport = self.get_transport()
        if transport.is_readonly():
            with open('a', 'w') as f:
                f.write('0123456789')
        else:
            transport.put_bytes('a', b'0123456789')

        d = list(transport.readv('a', ((0, 1),)))
        self.assertEqual(d[0], (0, b'0'))

        d = list(transport.readv('a', ((0, 1), (1, 1), (3, 2), (9, 1))))
        self.assertEqual(d[0], (0, b'0'))
        self.assertEqual(d[1], (1, b'1'))
        self.assertEqual(d[2], (3, b'34'))
        self.assertEqual(d[3], (9, b'9'))

    def test_readv_out_of_order(self):
        transport = self.get_transport()
        if transport.is_readonly():
            with open('a', 'w') as f:
                f.write('0123456789')
        else:
            transport.put_bytes('a', b'01234567890')

        d = list(transport.readv('a', ((1, 1), (9, 1), (0, 1), (3, 2))))
        self.assertEqual(d[0], (1, b'1'))
        self.assertEqual(d[1], (9, b'9'))
        self.assertEqual(d[2], (0, b'0'))
        self.assertEqual(d[3], (3, b'34'))

    def test_readv_with_adjust_for_latency(self):
        transport = self.get_transport()
        # the adjust for latency flag expands the data region returned
        # according to a per-transport heuristic, so testing is a little
        # tricky as we need more data than the largest combining that our
        # transports do. To accomodate this we generate random data and cross
        # reference the returned data with the random data. To avoid doing
        # multiple large random byte look ups we do several tests on the same
        # backing data.
        content = osutils.rand_bytes(200 * 1024)
        content_size = len(content)
        if transport.is_readonly():
            self.build_tree_contents([('a', content)])
        else:
            transport.put_bytes('a', content)

        def check_result_data(result_vector):
            for item in result_vector:
                data_len = len(item[1])
                self.assertEqual(content[item[0]:item[0] + data_len], item[1])

        # start corner case
        result = list(transport.readv('a', ((0, 30),),
                                      adjust_for_latency=True, upper_limit=content_size))
        # we expect 1 result, from 0, to something > 30
        self.assertEqual(1, len(result))
        self.assertEqual(0, result[0][0])
        self.assertTrue(len(result[0][1]) >= 30)
        check_result_data(result)
        # end of file corner case
        result = list(transport.readv('a', ((204700, 100),),
                                      adjust_for_latency=True, upper_limit=content_size))
        # we expect 1 result, from 204800- its length, to the end
        self.assertEqual(1, len(result))
        data_len = len(result[0][1])
        self.assertEqual(204800 - data_len, result[0][0])
        self.assertTrue(data_len >= 100)
        check_result_data(result)
        # out of order ranges are made in order
        result = list(transport.readv('a', ((204700, 100), (0, 50)),
                                      adjust_for_latency=True, upper_limit=content_size))
        # we expect 2 results, in order, start and end.
        self.assertEqual(2, len(result))
        # start
        data_len = len(result[0][1])
        self.assertEqual(0, result[0][0])
        self.assertTrue(data_len >= 30)
        # end
        data_len = len(result[1][1])
        self.assertEqual(204800 - data_len, result[1][0])
        self.assertTrue(data_len >= 100)
        check_result_data(result)
        # close ranges get combined (even if out of order)
        for request_vector in [((400, 50), (800, 234)), ((800, 234), (400, 50))]:
            result = list(transport.readv('a', request_vector,
                                          adjust_for_latency=True, upper_limit=content_size))
            self.assertEqual(1, len(result))
            data_len = len(result[0][1])
            # minimum length is from 400 to 1034 - 634
            self.assertTrue(data_len >= 634)
            # must contain the region 400 to 1034
            self.assertTrue(result[0][0] <= 400)
            self.assertTrue(result[0][0] + data_len >= 1034)
            check_result_data(result)

    def test_readv_with_adjust_for_latency_with_big_file(self):
        transport = self.get_transport()
        # test from observed failure case.
        if transport.is_readonly():
            with open('a', 'w') as f:
                f.write('a' * 1024 * 1024)
        else:
            transport.put_bytes('a', b'a' * 1024 * 1024)
        broken_vector = [(465219, 800), (225221, 800), (445548, 800),
                         (225037, 800), (221357, 800), (437077, 800), (947670, 800),
                         (465373, 800), (947422, 800)]
        results = list(transport.readv('a', broken_vector, True, 1024 * 1024))
        found_items = [False] * 9
        for pos, (start, length) in enumerate(broken_vector):
            # check the range is covered by the result
            for offset, data in results:
                if offset <= start and start + length <= offset + len(data):
                    found_items[pos] = True
        self.assertEqual([True] * 9, found_items)

    def test_get_with_open_write_stream_sees_all_content(self):
        t = self.get_transport()
        if t.is_readonly():
            return
        with t.open_write_stream('foo') as handle:
            handle.write(b'bcd')
            self.assertEqual([(0, b'b'), (2, b'd')], list(
                t.readv('foo', ((0, 1), (2, 1)))))

    def test_get_smart_medium(self):
        """All transports must either give a smart medium, or know they can't.
        """
        transport = self.get_transport()
        try:
            client_medium = transport.get_smart_medium()
        except errors.NoSmartMedium:
            # as long as we got it we're fine
            pass
        else:
            from ..bzr.smart import medium
            self.assertIsInstance(client_medium, medium.SmartClientMedium)

    def test_readv_short_read(self):
        transport = self.get_transport()
        if transport.is_readonly():
            with open('a', 'w') as f:
                f.write('0123456789')
        else:
            transport.put_bytes('a', b'01234567890')

        # This is intentionally reading off the end of the file
        # since we are sure that it cannot get there
        self.assertListRaises((errors.ShortReadvError, errors.InvalidRange,
                               # Can be raised by paramiko
                               AssertionError),
                              transport.readv, 'a', [(1, 1), (8, 10)])

        # This is trying to seek past the end of the file, it should
        # also raise a special error
        self.assertListRaises((errors.ShortReadvError, errors.InvalidRange),
                              transport.readv, 'a', [(12, 2)])

    def test_no_segment_parameters(self):
        """Segment parameters should be stripped and stored in
        transport.segment_parameters."""
        transport = self.get_transport("foo")
        self.assertEqual({}, transport.get_segment_parameters())

    def test_segment_parameters(self):
        """Segment parameters should be stripped and stored in
        transport.get_segment_parameters()."""
        base_url = self._server.get_url()
        parameters = {"key1": "val1", "key2": "val2"}
        url = urlutils.join_segment_parameters(base_url, parameters)
        transport = _mod_transport.get_transport_from_url(url)
        self.assertEqual(parameters, transport.get_segment_parameters())

    def test_set_segment_parameters(self):
        """Segment parameters can be set and show up in base."""
        transport = self.get_transport("foo")
        orig_base = transport.base
        transport.set_segment_parameter("arm", "board")
        self.assertEqual("%s,arm=board" % orig_base, transport.base)
        self.assertEqual({"arm": "board"}, transport.get_segment_parameters())
        transport.set_segment_parameter("arm", None)
        transport.set_segment_parameter("nonexistant", None)
        self.assertEqual({}, transport.get_segment_parameters())
        self.assertEqual(orig_base, transport.base)

    def test_stat_symlink(self):
        # if a transport points directly to a symlink (and supports symlinks
        # at all) you can tell this.  helps with bug 32669.
        t = self.get_transport()
        try:
            t.symlink('target', 'link')
        except TransportNotPossible:
            raise TestSkipped("symlinks not supported")
        t2 = t.clone('link')
        st = t2.stat('')
        self.assertTrue(stat.S_ISLNK(st.st_mode))

    def test_abspath_url_unquote_unreserved(self):
        """URLs from abspath should have unreserved characters unquoted

        Need consistent quoting notably for tildes, see lp:842223 for more.
        """
        t = self.get_transport()
        needlessly_escaped_dir = "%2D%2E%30%39%41%5A%5F%61%7A%7E/"
        self.assertEqual(t.base + "-.09AZ_az~",
                         t.abspath(needlessly_escaped_dir))

    def test_clone_url_unquote_unreserved(self):
        """Base URL of a cloned branch needs unreserved characters unquoted

        Cloned transports should be prefix comparable for things like the
        isolation checking of tests, see lp:842223 for more.
        """
        t1 = self.get_transport()
        needlessly_escaped_dir = "%2D%2E%30%39%41%5A%5F%61%7A%7E/"
        self.build_tree([needlessly_escaped_dir], transport=t1)
        t2 = t1.clone(needlessly_escaped_dir)
        self.assertEqual(t1.base + "-.09AZ_az~/", t2.base)

    def test_hook_post_connection_one(self):
        """Fire post_connect hook after a ConnectedTransport is first used"""
        log = []
        Transport.hooks.install_named_hook("post_connect", log.append, None)
        t = self.get_transport()
        self.assertEqual([], log)
        t.has("non-existant")
        if isinstance(t, RemoteTransport):
            self.assertEqual([t.get_smart_medium()], log)
        elif isinstance(t, ConnectedTransport):
            self.assertEqual([t], log)
        else:
            self.assertEqual([], log)

    def test_hook_post_connection_multi(self):
        """Fire post_connect hook once per unshared underlying connection"""
        log = []
        Transport.hooks.install_named_hook("post_connect", log.append, None)
        t1 = self.get_transport()
        t2 = t1.clone(".")
        t3 = self.get_transport()
        self.assertEqual([], log)
        t1.has("x")
        t2.has("x")
        t3.has("x")
        if isinstance(t1, RemoteTransport):
            self.assertEqual([t.get_smart_medium() for t in [t1, t3]], log)
        elif isinstance(t1, ConnectedTransport):
            self.assertEqual([t1, t3], log)
        else:
            self.assertEqual([], log)



# Copyright (C) 2006, 2007, 2009, 2010, 2011, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Black-box tests for brz revert."""

import os

import breezy.osutils
from breezy.tests import TestCaseWithTransport
from breezy.trace import mutter
from breezy.workingtree import WorkingTree


class TestRevert(TestCaseWithTransport):

    def _prepare_tree(self):
        self.run_bzr('init')
        self.run_bzr('mkdir dir')

        with open('dir/file', 'wb') as f:
            f.write(b'spam')
        self.run_bzr('add dir/file')

        self.run_bzr('commit -m1')

        # modify file
        with open('dir/file', 'wb') as f:
            f.write(b'eggs')

        # check status
        self.assertEqual('modified:\n  dir/file\n', self.run_bzr('status')[0])

    def _prepare_rename_mod_tree(self):
        self.build_tree(['a/', 'a/b', 'a/c', 'a/d/', 'a/d/e', 'f/', 'f/g',
                         'f/h', 'f/i'])
        self.run_bzr('init')
        self.run_bzr('add')
        self.run_bzr('commit -m 1')
        wt = WorkingTree.open('.')
        wt.rename_one('a/b', 'f/b')
        wt.rename_one('a/d/e', 'f/e')
        wt.rename_one('a/d', 'f/d')
        wt.rename_one('f/g', 'a/g')
        wt.rename_one('f/h', 'h')
        wt.rename_one('f', 'j')

    def helper(self, param=''):
        self._prepare_tree()
        # change dir
        # revert to default revision for file in subdir does work
        os.chdir('dir')
        mutter('cd dir\n')

        self.assertEqual('1\n', self.run_bzr('revno')[0])
        self.run_bzr('revert %s file' % param)
        with open('file', 'rb') as f:
            self.assertEqual(b'spam', f.read())

    def test_revert_in_subdir(self):
        self.helper()

    def test_revert_to_revision_in_subdir(self):
        # test case for bug #29424:
        # revert to specific revision for file in subdir does not work
        self.helper('-r 1')

    def test_revert_in_checkout(self):
        os.mkdir('brach')
        os.chdir('brach')
        self._prepare_tree()
        self.run_bzr('checkout --lightweight . ../sprach')
        self.run_bzr('commit -m more')
        os.chdir('../sprach')
        self.assertEqual('', self.run_bzr('status')[0])
        self.run_bzr('revert')
        self.assertEqual('', self.run_bzr('status')[0])

    def test_revert_dirname(self):
        """Test that revert DIRECTORY does what's expected"""
        self._prepare_rename_mod_tree()
        self.run_bzr('revert a')
        self.assertPathExists('a/b')
        self.assertPathExists('a/d')
        self.assertPathDoesNotExist('a/g')
        self.expectFailure(
            "j is in the delta revert applies because j was renamed too",
            self.assertPathExists, 'j')
        self.assertPathExists('h')
        self.run_bzr('revert f')
        self.assertPathDoesNotExist('j')
        self.assertPathDoesNotExist('h')
        self.assertPathExists('a/d/e')

    def test_revert_chatter(self):
        self._prepare_rename_mod_tree()
        chatter = self.run_bzr('revert')[1]
        self.assertEqualDiff(
            'R   a/g => f/g\n'
            'R   h => f/h\n'
            'R   j/ => f/\n'
            'R   j/b => a/b\n'
            'R   j/d/ => a/d/\n'
            'R   j/e => a/d/e\n',
            chatter)

    def test_revert(self):
        self.run_bzr('init')

        with open('hello', 'w') as f:
            f.write('foo')
        self.run_bzr('add hello')
        self.run_bzr('commit -m setup hello')

        with open('goodbye', 'w') as f:
            f.write('baz')
        self.run_bzr('add goodbye')
        self.run_bzr('commit -m setup goodbye')

        with open('hello', 'w') as f:
            f.write('bar')
        with open('goodbye', 'w') as f:
            f.write('qux')
        self.run_bzr('revert hello')
        self.check_file_contents('hello', b'foo')
        self.check_file_contents('goodbye', b'qux')
        self.run_bzr('revert')
        self.check_file_contents('goodbye', b'baz')

        os.mkdir('revertdir')
        self.run_bzr('add revertdir')
        self.run_bzr('commit -m f')
        os.rmdir('revertdir')
        self.run_bzr('revert')

        if breezy.osutils.supports_symlinks(self.test_dir):
            os.symlink('/unlikely/to/exist', 'symlink')
            self.run_bzr('add symlink')
            self.run_bzr('commit -m f')
            os.unlink('symlink')
            self.run_bzr('revert')
            self.assertPathExists('symlink')
            os.unlink('symlink')
            os.symlink('a-different-path', 'symlink')
            self.run_bzr('revert')
            self.assertEqual('/unlikely/to/exist',
                             os.readlink('symlink'))
        else:
            self.log("skipping revert symlink tests")

        with open('hello', 'w') as f:
            f.write('xyz')
        self.run_bzr('commit -m xyz hello')
        self.run_bzr('revert -r 1 hello')
        self.check_file_contents('hello', b'foo')
        self.run_bzr('revert hello')
        self.check_file_contents('hello', b'xyz')
        os.chdir('revertdir')
        self.run_bzr('revert')
        os.chdir('..')

    def test_revert_newly_added(self):
        # this tests the UI reports reverting a newly added file
        # correct (such files are not deleted)
        tree = self.make_branch_and_tree('.')
        self.build_tree(['file'])
        tree.add(['file'])
        out, err = self.run_bzr('revert')
        self.assertEqual('', out)
        self.assertEqual('-   file\n', err)

    def test_revert_removing_file(self):
        # this tests the UI reports reverting a file which has been committed
        # to a revision that did not have it, reports it as being deleted.
        tree = self.make_branch_and_tree('.')
        tree.commit('empty commit')
        self.build_tree(['file'])
        tree.add(['file'])
        tree.commit('add file')
        out, err = self.run_bzr('revert -r -2')
        self.assertEqual('', out)
        self.assertEqual('-D  file\n', err)

    def test_revert_forget_merges(self):
        # revert --forget-merges removes any pending merges into the tree, but
        # leaves the files unchanged
        tree = self.make_branch_and_tree('.')
        # forget-merges before first commit, though pointless, does not fail
        self.run_bzr(['revert', '--forget-merges'])
        self.build_tree(['file'])
        first_rev_id = tree.commit('initial commit')
        self.build_tree_contents([('file', b'new content')])
        existing_parents = tree.get_parent_ids()
        self.assertEqual([first_rev_id], existing_parents)
        merged_parents = existing_parents + [b'merged-in-rev']
        tree.set_parent_ids(merged_parents)
        self.assertEqual(merged_parents, tree.get_parent_ids())
        self.run_bzr(['revert', '--forget-merges'])
        self.assertEqual([first_rev_id], tree.get_parent_ids())
        # changed files are not reverted
        self.assertFileEqual(b'new content', 'file')
        # you can give it the path of a tree
        self.run_bzr(['revert', '--forget-merges', tree.abspath('.')])



# Copyright (C) 2006, 2007, 2009, 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for lock-breaking user interface"""

from breezy import branch, config, controldir, errors, osutils, tests
from breezy.tests.script import run_script


class TestBreakLock(tests.TestCaseWithTransport):

    # General principal for break-lock: All the elements that might be locked
    # by a brz operation on PATH, are candidates that break-lock may unlock.
    # so pathologically if we have a lightweight checkout A, of branch B, which
    # is bound to location C, the following things should be checked for locks
    # to break:
    # wt = WorkingTree(A)
    # wt.branch
    # wt.branch.repository
    # wt.branch.get_master_branch()
    # wt.branch.get_master_branch().repository
    # so for smoke tests all we need is a bound branch with a checkout of that
    # and we can then use different urls to test individual cases, for as much
    # granularity as needed.

    def setUp(self):
        super().setUp()
        self.build_tree(
            ['master-repo/',
             'master-repo/master-branch/',
             'repo/',
             'repo/branch/',
             'checkout/'])
        controldir.ControlDir.create('master-repo').create_repository()
        self.master_branch = controldir.ControlDir.create_branch_convenience(
            'master-repo/master-branch')
        controldir.ControlDir.create('repo').create_repository()
        local_branch = controldir.ControlDir.create_branch_convenience(
            'repo/branch')
        try:
            local_branch.bind(self.master_branch)
        except branch.BindingUnsupported:
            raise tests.TestNotApplicable(
                'default format does not support bound branches')
        checkoutdir = controldir.ControlDir.create('checkout')
        checkoutdir.set_branch_reference(local_branch)
        self.wt = checkoutdir.create_workingtree()

    def test_break_lock_help(self):
        out, err = self.run_bzr('break-lock --help')
        # shouldn't fail and should not produce error output
        self.assertEqual('', err)

    def test_break_lock_no_interaction(self):
        """With --force, the user isn't asked for confirmation"""
        self.master_branch.lock_write()
        run_script(self, """
        $ brz break-lock --force master-repo/master-branch
        Broke lock ...master-branch/.bzr/...
        """)
        # lock should now be dead
        self.assertRaises(errors.LockBroken, self.master_branch.unlock)

    def test_break_lock_everything_locked(self):
        # if everything is locked, we should be able to unlock the lot.
        # however, we dont test breaking the working tree because we
        # cannot accurately do so right now: the dirstate lock is held
        # by an os lock, and we need to spawn a separate process to lock it
        # then kill -9 it.
        # sketch of test:
        # lock most of the dir:
        self.wt.branch.lock_write()
        self.master_branch.lock_write()
        # run the break-lock
        # we need 5 yes's - wt, branch, repo, bound branch, bound repo.
        self.run_bzr('break-lock checkout', stdin="y\ny\ny\ny\n")
        # a new tree instance should be lockable
        br = branch.Branch.open('checkout')
        br.lock_write()
        br.unlock()
        # and a new instance of the master branch
        mb = br.get_master_branch()
        mb.lock_write()
        mb.unlock()
        self.assertRaises(errors.LockBroken, self.wt.unlock)
        self.assertRaises(errors.LockBroken, self.master_branch.unlock)


class TestConfigBreakLock(tests.TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        self.config_file_name = './my.conf'
        self.build_tree_contents([(self.config_file_name,
                                   b'[DEFAULT]\none=1\n')])
        self.config = config.LockableConfig(file_name=self.config_file_name)
        self.config.lock_write()

    def test_create_pending_lock(self):
        self.addCleanup(self.config.unlock)
        self.assertTrue(self.config._lock.is_held)

    def test_break_lock(self):
        self.run_bzr('break-lock --config %s'
                     % osutils.dirname(self.config_file_name),
                     stdin="y\n")
        self.assertRaises(errors.LockBroken, self.config.unlock)



# Copyright (C) 2007 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from .knit import KnitCorrupt


def _load_data_py(kndx, fp):
    """Read in a knit index."""
    cache = kndx._cache
    history = kndx._history

    kndx.check_header(fp)
    # readlines reads the whole file at once:
    # bad for transports like http, good for local disk
    # we save 60 ms doing this one change (
    # from calling readline each time to calling
    # readlines once.
    # probably what we want for nice behaviour on
    # http is a incremental readlines that yields, or
    # a check for local vs non local indexes,
    history_top = len(history) - 1
    for line in fp.readlines():
        rec = line.split()
        if len(rec) < 5 or rec[-1] != b':':
            # corrupt line.
            # FIXME: in the future we should determine if its a
            # short write - and ignore it
            # or a different failure, and raise. RBC 20060407
            continue

        try:
            parents = []
            for value in rec[4:-1]:
                if value[:1] == b'.':
                    # uncompressed reference
                    parent_id = value[1:]
                else:
                    parent_id = history[int(value)]
                parents.append(parent_id)
        except (IndexError, ValueError) as e:
            # The parent could not be decoded to get its parent row. This
            # at a minimum will cause this row to have wrong parents, or
            # even to apply a delta to the wrong base and decode
            # incorrectly. its therefore not usable, and because we have
            # encountered a situation where a new knit index had this
            # corrupt we can't asssume that no other rows referring to the
            # index of this record actually mean the subsequent uncorrupt
            # one, so we error.
            raise KnitCorrupt(kndx._filename, "line {!r}: {}".format(rec, e))

        version_id, options, pos, size = rec[:4]
        try:
            pos = int(pos)
        except ValueError as e:
            raise KnitCorrupt(kndx._filename,
                              "invalid position on line %r: %s"
                              % (rec, e))
        try:
            size = int(size)
        except ValueError as e:
            raise KnitCorrupt(kndx._filename,
                              "invalid size on line %r: %s"
                              % (rec, e))

        # See kndx._cache_version
        # only want the _history index to reference the 1st
        # index entry for version_id
        if version_id not in cache:
            history_top += 1
            index = history_top
            history.append(version_id)
        else:
            index = cache[version_id][5]
        cache[version_id] = (version_id,
                             options.split(b','),
                             pos,
                             size,
                             tuple(parents),
                             index)
        # end kndx._cache_version



# Copyright (C) 2020 Jelmer Vernooij <jelmer@jelmer.uk>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""GitLab command implementations."""

from ... import errors, urlutils
from ...commands import Command
from ...option import Option
from ...trace import note


class cmd_gitlab_login(Command):
    __doc__ = """Log into a GitLab instance.

    This command takes a GitLab instance URL (e.g. https://gitlab.com)
    as well as an optional private token. Private tokens can be created via the
    web UI.

    :Examples:

      Log into GNOME's GitLab (prompts for a token):

         brz gitlab-login https://gitlab.gnome.org/

      Log into Debian's salsa, using a token created earlier:

         brz gitlab-login https://salsa.debian.org if4Theis6Eich7aef0zo
    """

    takes_args = ['url', 'private_token?']

    takes_options = [
        Option('name', help='Name for GitLab site in configuration.',
               type=str),
        Option('no-check',
               "Don't check that the token is valid."),
        ]

    def run(self, url, private_token=None, name=None, no_check=False):
        from breezy import ui

        from .forge import store_gitlab_token
        if name is None:
            try:
                name = urlutils.parse_url(url)[3].split('.')[-2]
            except (ValueError, IndexError):
                raise errors.CommandError(
                    'please specify a site name with --name')
        if private_token is None:
            note("Please visit %s to obtain a private token.",
                 urlutils.join(url, "-/profile/personal_access_tokens"))
            private_token = ui.ui_factory.get_password('Private token')
        if not no_check:
            from breezy.transport import get_transport

            from .forge import GitLab
            GitLab(get_transport(url), private_token=private_token)
        store_gitlab_token(name=name, url=url, private_token=private_token)



# Copyright (C) 2007-2010, 2016 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


"""Black-box tests for 'brz unknowns', which shows unknown files."""

from breezy.tests import TestCaseWithTransport


class TestUnknowns(TestCaseWithTransport):

    def test_unknowns(self):
        """Test that 'unknown' command reports unknown files"""

        # in empty directory, no unknowns
        tree = self.make_branch_and_tree('.')
        self.assertEqual(self.run_bzr('unknowns')[0], '')

        # single unknown file
        self.build_tree_contents([('a', b'contents of a\n')])
        self.assertEqual(self.run_bzr('unknowns')[0], 'a\n')

        # multiple unknown files, including one with a space in its name
        self.build_tree(['b', 'c', 'd e'])
        self.assertEqual(self.run_bzr('unknowns')[0], 'a\nb\nc\n"d e"\n')

        # after add, file no longer shown
        tree.add(['a', 'd e'])
        self.assertEqual(self.run_bzr('unknowns')[0], 'b\nc\n')

        # after all added, none shown
        tree.add(['b', 'c'])
        self.assertEqual(self.run_bzr('unknowns')[0], '')

    def test_unknowns_directory(self):
        """Test --directory option"""
        tree = self.make_branch_and_tree('a')
        self.build_tree(['a/README'])
        out, err = self.run_bzr(['unknowns', '--directory=a'])
        self.assertEqual('README\n', out)



# Copyright (C) 2010 Canonical Ltd
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Matchers for breezy.

Primarily test support, Matchers are used by self.assertThat in the breezy
test suite. A matcher is a stateful test helper which can be used to determine
if a passed object 'matches', much like a regex. If the object does not match
the mismatch can be described in a human readable fashion. assertThat then
raises if a mismatch occurs, showing the description as the assertion error.

Matchers are designed to be more reusable and composable than layered
assertions in Test Case objects, so they are recommended for new testing work.
"""

__all__ = [
    'HasLayout',
    'HasPathRelations',
    'MatchesAncestry',
    'ReturnsUnlockable',
    'RevisionHistoryMatches',
    'MatchesTreeChanges',
    ]

from testtools.matchers import Equals, Matcher, Mismatch

from .. import osutils
from .. import revision as _mod_revision
from ..tree import InterTree, TreeChange


class ReturnsUnlockable(Matcher):
    """A matcher that checks for the pattern we want lock* methods to have:

    They should return an object with an unlock() method.
    Calling that method should unlock the original object.

    :ivar lockable_thing: The object which can be locked that will be
        inspected.
    """

    def __init__(self, lockable_thing):
        Matcher.__init__(self)
        self.lockable_thing = lockable_thing

    def __str__(self):
        return ('ReturnsUnlockable(lockable_thing=%s)' %
                self.lockable_thing)

    def match(self, lock_method):
        lock_method().unlock()
        if self.lockable_thing.is_locked():
            return _IsLocked(self.lockable_thing)
        return None


class _IsLocked(Mismatch):
    """Something is locked."""

    def __init__(self, lockable_thing):
        self.lockable_thing = lockable_thing

    def describe(self):
        return "%s is locked" % self.lockable_thing


class _AncestryMismatch(Mismatch):
    """Ancestry matching mismatch."""

    def __init__(self, tip_revision, got, expected):
        self.tip_revision = tip_revision
        self.got = got
        self.expected = expected

    def describe(self):
        return "mismatched ancestry for revision {!r} was {!r}, expected {!r}".format(
            self.tip_revision, self.got, self.expected)


class MatchesAncestry(Matcher):
    """A matcher that checks the ancestry of a particular revision.

    :ivar graph: Graph in which to check the ancestry
    :ivar revision_id: Revision id of the revision
    """

    def __init__(self, repository, revision_id):
        Matcher.__init__(self)
        self.repository = repository
        self.revision_id = revision_id

    def __str__(self):
        return ('MatchesAncestry(repository={!r}, revision_id={!r})'.format(
            self.repository, self.revision_id))

    def match(self, expected):
        with self.repository.lock_read():
            graph = self.repository.get_graph()
            got = [r for r, p in graph.iter_ancestry([self.revision_id])]
            if _mod_revision.NULL_REVISION in got:
                got.remove(_mod_revision.NULL_REVISION)
        if sorted(got) != sorted(expected):
            return _AncestryMismatch(self.revision_id, sorted(got),
                                     sorted(expected))


class HasLayout(Matcher):
    """A matcher that checks if a tree has a specific layout.

    :ivar entries: List of expected entries, as (path, file_id) pairs.
    """

    def __init__(self, entries):
        Matcher.__init__(self)
        self.entries = entries

    def get_tree_layout(self, tree, include_file_ids):
        """Get the (path, file_id) pairs for the current tree."""
        with tree.lock_read():
            for path, ie in tree.iter_entries_by_dir():
                if path != '':
                    path += ie.kind_character()
                if include_file_ids:
                    yield (path, ie.file_id)
                else:
                    yield path

    @staticmethod
    def _strip_unreferenced_directories(entries):
        """Strip all directories that don't (in)directly contain any files.

        :param entries: List of path strings or (path, ie) tuples to process
        """
        directories = []
        for entry in entries:
            if isinstance(entry, str):
                path = entry
            else:
                path = entry[0]
            if not path or path[-1] == "/":
                # directory
                directories.append((path, entry))
            else:
                # Yield the referenced parent directories
                for dirpath, direntry in directories:
                    if osutils.is_inside(dirpath, path):
                        yield direntry
                directories = []
                yield entry

    def __str__(self):
        return 'HasLayout(%r)' % self.entries

    def match(self, tree):
        include_file_ids = self.entries and not isinstance(
            self.entries[0], str)
        actual = list(self.get_tree_layout(
            tree, include_file_ids=include_file_ids))
        if not tree.has_versioned_directories():
            entries = list(self._strip_unreferenced_directories(self.entries))
        else:
            entries = self.entries
        return Equals(entries).match(actual)


class HasPathRelations(Matcher):
    """Matcher verifies that paths have a relation to those in another tree.

    :ivar previous_tree: tree to compare to
    :ivar previous_entries: List of expected entries, as (path, previous_path) pairs.
    """

    def __init__(self, previous_tree, previous_entries):
        Matcher.__init__(self)
        self.previous_tree = previous_tree
        self.previous_entries = previous_entries

    def get_path_map(self, tree):
        """Get the (path, previous_path) pairs for the current tree."""
        previous_intertree = InterTree.get(self.previous_tree, tree)
        with tree.lock_read(), self.previous_tree.lock_read():
            for path, ie in tree.iter_entries_by_dir():
                if tree.supports_rename_tracking():
                    previous_path = previous_intertree.find_source_path(path)
                else:
                    if self.previous_tree.is_versioned(path):
                        previous_path = path
                    else:
                        previous_path = None
                if previous_path:
                    kind = self.previous_tree.kind(previous_path)
                    if kind == 'directory':
                        previous_path += '/'
                if path == '':
                    yield ("", previous_path)
                else:
                    yield (path + ie.kind_character(), previous_path)

    @staticmethod
    def _strip_unreferenced_directories(entries):
        """Strip all directories that don't (in)directly contain any files.

        :param entries: List of path strings or (path, previous_path) tuples to process
        """
        directory_used = set()
        directories = []
        for (path, previous_path) in entries:
            if not path or path[-1] == "/":
                # directory
                directories.append((path, previous_path))
            else:
                # Yield the referenced parent directories
                for direntry in directories:
                    if osutils.is_inside(direntry[0], path):
                        directory_used.add(direntry[0])
        for (path, previous_path) in entries:
            if (not path.endswith("/")) or path in directory_used:
                yield (path, previous_path)

    def __str__(self):
        return 'HasPathRelations({!r}, {!r})'.format(self.previous_tree, self.previous_entries)

    def match(self, tree):
        actual = list(self.get_path_map(tree))
        if not tree.has_versioned_directories():
            entries = list(self._strip_unreferenced_directories(
                self.previous_entries))
        else:
            entries = self.previous_entries
        if not tree.supports_rename_tracking():
            entries = [
                (path, path if self.previous_tree.is_versioned(path) else None)
                for (path, previous_path) in entries]
        return Equals(entries).match(actual)


class RevisionHistoryMatches(Matcher):
    """A matcher that checks if a branch has a specific revision history.

    :ivar history: Revision history, as list of revisions. Oldest first.
    """

    def __init__(self, history):
        Matcher.__init__(self)
        self.expected = history

    def __str__(self):
        return 'RevisionHistoryMatches(%r)' % self.expected

    def match(self, branch):
        with branch.lock_read():
            graph = branch.repository.get_graph()
            history = list(graph.iter_lefthand_ancestry(
                branch.last_revision(), [_mod_revision.NULL_REVISION]))
            history.reverse()
        return Equals(self.expected).match(history)


class MatchesTreeChanges(Matcher):
    """A matcher that checks that tree changes match expected contents."""

    def __init__(self, old_tree, new_tree, expected):
        Matcher.__init__(self)
        expected = [TreeChange(*x) if isinstance(x, tuple) else x for x in expected]
        self.use_inventory_tree_changes = old_tree.supports_file_ids and new_tree.supports_file_ids
        self.expected = expected
        self.old_tree = old_tree
        self.new_tree = new_tree

    @staticmethod
    def _convert_to_inventory_tree_changes(old_tree, new_tree, expected):
        from ..bzr.inventorytree import InventoryTreeChange
        rich_expected = []

        def get_parent_id(t, p):
            if p:
                return t.path2id(osutils.dirname(p))
            else:
                return None

        for c in expected:
            if c.path[0] is not None:
                file_id = old_tree.path2id(c.path[0])
            else:
                file_id = new_tree.path2id(c.path[1])
            old_parent_id = get_parent_id(old_tree, c.path[0])
            new_parent_id = get_parent_id(new_tree, c.path[1])
            rich_expected.append(
                InventoryTreeChange(
                    file_id=file_id,
                    parent_id=(old_parent_id, new_parent_id),
                    path=c.path, changed_content=c.changed_content,
                    versioned=c.versioned, name=c.name,
                    kind=c.kind, executable=c.executable,
                    copied=c.copied))
        return rich_expected

    def __str__(self):
        return '<MatchesTreeChanges(%r)>' % self.expected

    def match(self, actual):
        from ..bzr.inventorytree import InventoryTreeChange
        actual = list(actual)
        if self.use_inventory_tree_changes or (actual and isinstance(actual[0], InventoryTreeChange)):
            expected = self._convert_to_inventory_tree_changes(self.old_tree, self.new_tree, self.expected)
        else:
            expected = self.expected
        if self.use_inventory_tree_changes:
            actual = self._convert_to_inventory_tree_changes(self.old_tree, self.new_tree, actual)
        return Equals(expected).match(actual)



# Copyright (C) 2005 Canonical Ltd
#   Authors: Robert Collins <robert.collins@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""This module provides a transactional facility.

Transactions provide hooks to allow data objects (i.e. inventory weaves or
the revision-history file) to be placed in a registry and retrieved later
during the same transaction. Transactions in bzr are not atomic - they
depend on data ordering of writes, so we do not have commit or rollback
facilities at the transaction level.

Read only transactions raise an assert when objects are listed as dirty
against them - preventing unintended writes.

Write transactions preserve dirty objects in the cache, though due to the
write ordering approach we use for consistency 'dirty' is a misleading term.
A dirty object is one we have modified.

Both read and write transactions *may* flush unchanged objects out of
memory, unless they are marked as 'precious' which indicates that
repeated reads cannot be obtained if the object is ejected, or that
the object is an expensive one for obtaining.
"""

import sys

from . import errors as errors
from .identitymap import IdentityMap, NullIdentityMap
from .trace import mutter


class Transaction:
    """Base class for transactions."""

    def writeable(self) -> bool:
        raise NotImplementedError(self.writeable)


class ReadOnlyTransaction(Transaction):
    """A read only unit of work for data objects."""

    def finish(self):
        """Clean up this transaction."""

    def __init__(self):
        super().__init__()
        self.map = IdentityMap()
        self._clean_objects = set()
        self._clean_queue = []
        self._limit = -1
        self._precious_objects = set()

    def is_clean(self, an_object):
        """Return True if an_object is clean."""
        return (an_object in self._clean_objects)

    def register_clean(self, an_object, precious=False):
        """Register an_object as being clean.

        If the precious hint is True, the object will not
        be ejected from the object identity map ever.
        """
        self._clean_objects.add(an_object)
        self._clean_queue.append(an_object)
        if precious:
            self._precious_objects.add(an_object)
        self._trim()

    def register_dirty(self, an_object):
        """Register an_object as being dirty."""
        raise errors.ReadOnlyObjectDirtiedError(an_object)

    def set_cache_size(self, size):
        """Set a new cache size."""
        if size < -1:
            raise ValueError(size)
        self._limit = size
        self._trim()

    def _trim(self):
        """Trim the cache back if needed."""
        if self._limit < 0 or self._limit - len(self._clean_objects) > 0:
            return
        needed = len(self._clean_objects) - self._limit
        offset = 0
        while needed and offset < len(self._clean_objects):
            # references we know of:
            # temp passed to getrefcount in our frame
            # temp in getrefcount's frame
            # the map forward
            # the map backwards
            # _clean_objects
            # _clean_queue
            if sys.version_info >= (3, 11):
                ref_threshold = 6
            else:
                # 1 missing on Python < 3.11
                ref_threshold = 7
            if (sys.getrefcount(self._clean_queue[offset]) <= ref_threshold
                    and not self._clean_queue[offset] in self._precious_objects):
                removed = self._clean_queue[offset]
                self._clean_objects.remove(removed)
                del self._clean_queue[offset]
                self.map.remove_object(removed)
                mutter('removed object %r', removed)
                needed -= 1
            else:
                offset += 1

    def writeable(self):
        """Read only transactions do not allow writes."""
        return False


class WriteTransaction(ReadOnlyTransaction):
    """A write transaction

    - caches domain objects
    - clean objects can be removed from the cache
    - dirty objects are retained.
    """

    def finish(self):
        """Clean up this transaction."""
        for thing in self._dirty_objects:
            callback = getattr(thing, 'transaction_finished', None)
            if callback is not None:
                callback()

    def __init__(self):
        super().__init__()
        self._dirty_objects = set()

    def is_dirty(self, an_object):
        """Return True if an_object is dirty."""
        return (an_object in self._dirty_objects)

    def register_dirty(self, an_object):
        """Register an_object as being dirty.

        Dirty objects are not ejected from the identity map
        until the transaction finishes and get informed
        when the transaction finishes.
        """
        self._dirty_objects.add(an_object)
        if self.is_clean(an_object):
            self._clean_objects.remove(an_object)
            del self._clean_queue[self._clean_queue.index(an_object)]
        self._trim()

    def writeable(self):
        """Write transactions allow writes."""
        return True


class PassThroughTransaction(Transaction):
    """A pass through transaction

    - nothing is cached.
    - nothing ever gets into the identity map.
    """

    def finish(self):
        """Clean up this transaction."""
        for thing in self._dirty_objects:
            callback = getattr(thing, 'transaction_finished', None)
            if callback is not None:
                callback()

    def __init__(self):
        super().__init__()
        self.map = NullIdentityMap()
        self._dirty_objects = set()

    def register_clean(self, an_object, precious=False):
        """Register an_object as being clean.

        Note that precious is only a hint, and PassThroughTransaction
        ignores it.
        """

    def register_dirty(self, an_object):
        """Register an_object as being dirty.

        Dirty objects get informed
        when the transaction finishes.
        """
        self._dirty_objects.add(an_object)

    def set_cache_size(self, ignored):
        """Do nothing, we are passing through."""

    def writeable(self):
        """Pass through transactions allow writes."""
        return True



# Copyright (C) 2006, 2008-2012, 2016 Canonical Ltd
# Authors:  Robert Collins <robert.collins@canonical.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""Tests for the RevisionTree class."""

from breezy import revision
from breezy.tests import TestCaseWithTransport
from breezy.tree import FileTimestampUnavailable


class TestTreeWithCommits(TestCaseWithTransport):

    def setUp(self):
        super().setUp()
        self.t = self.make_branch_and_tree('.')
        self.rev_id = self.t.commit('foo', allow_pointless=True)
        self.rev_tree = self.t.branch.repository.revision_tree(self.rev_id)

    def test_empty_no_unknowns(self):
        self.assertEqual([], list(self.rev_tree.unknowns()))

    def test_no_conflicts(self):
        self.assertEqual([], list(self.rev_tree.conflicts()))

    def test_parents(self):
        """RevisionTree.parent_ids should match the revision graph."""
        # XXX: TODO: Should this be a repository_implementation test ?
        # at the end of the graph, we get []
        self.assertEqual([], self.rev_tree.get_parent_ids())
        # do a commit to look further up
        revid_2 = self.t.commit('bar', allow_pointless=True)
        self.assertEqual(
            [self.rev_id],
            self.t.branch.repository.revision_tree(revid_2).get_parent_ids())
        # TODO commit a merge and check it is reported correctly.

        # the parents for a revision_tree(NULL_REVISION) are []:
        self.assertEqual([],
                         self.t.branch.repository.revision_tree(
            revision.NULL_REVISION).get_parent_ids())

    def test_empty_no_root(self):
        null_tree = self.t.branch.repository.revision_tree(
            revision.NULL_REVISION)
        self.assertIs(None, null_tree.path2id(''))

    def test_get_file_revision_root(self):
        self.assertEqual(self.rev_id, self.rev_tree.get_file_revision(''))

    def test_get_file_revision(self):
        self.build_tree_contents([('a', b'initial')])
        self.t.add(['a'])
        revid1 = self.t.commit('add a')
        revid2 = self.t.commit('another change', allow_pointless=True)
        tree = self.t.branch.repository.revision_tree(revid2)
        self.assertEqual(revid1,
                         tree.get_file_revision('a'))

    def test_get_file_mtime_ghost(self):
        path = next(iter(self.rev_tree.all_versioned_paths()))
        self.rev_tree.root_inventory.get_entry(
            self.rev_tree.path2id(path)).revision = b'ghostrev'
        self.assertRaises(FileTimestampUnavailable,
                          self.rev_tree.get_file_mtime, path)
